@@ -13,7 +13,7 @@ LaunchCLIOptions = AskCLIOptions = ClipboardCLIOptions = DiffCLIOptions = CLIOpt
 HintsCLIOptions = IcatCLIOptions = PanelCLIOptions = ResizeCLIOptions = CLIOptions
 ErrorCLIOptions = UnicodeCLIOptions = RCOptions = RemoteFileCLIOptions = CLIOptions
 QueryTerminalCLIOptions = BroadcastCLIOptions = ShowKeyCLIOptions = CLIOptions
-ThemesCLIOptions = TransferCLIOptions = CLIOptions
+ThemesCLIOptions = TransferCLIOptions = CopyCLIOptions = CLIOptions
 
 
 def generate_stub() -> None:
@@ -78,6 +78,9 @@ def generate_stub() -> None:
     from kittens.transfer.main import option_text as OPTIONS
     do(OPTIONS(), 'TransferCLIOptions')
 
+    from kittens.ssh.copy import option_text as OPTIONS
+    do(OPTIONS(), 'CopyCLIOptions')
+
     from kitty.rc.base import all_command_names, command_for_name
     for cmd_name in all_command_names():
         cmd = command_for_name(cmd_name)
