@@ -14,10 +14,10 @@ from kittens.runner import (
 )
 
 from .cli import (
-    OptionDict, OptionSpecSeq, options_for_completion, parse_option_spec,
-    prettify
+    OptionDict, options_for_completion, parse_option_spec, prettify
 )
-from .constants import shell_integration_dir
+from .remote_control import global_options_spec
+from .constants import config_dir, shell_integration_dir
 from .fast_data_types import truncate_point_for_length, wcswidth
 from .rc.base import all_command_names, command_for_name
 from .shell import options_for_cmd
@@ -278,14 +278,17 @@ def zsh_output_serializer(ans: Completions) -> str:
             allm = tuple(matches)
             if len(allm) > 1:
                 common_prefix = os.path.commonprefix(allm)
-                if common_prefix and '/' in common_prefix:
-                    common_prefix = os.path.dirname(common_prefix) + '/'
+                if common_prefix and os.sep in common_prefix:
+                    common_prefix = os.path.dirname(common_prefix).rstrip(os.sep) + os.sep
                     cmd.extend(('-p', shlex.quote(common_prefix)))
                     matches = MatchGroup({k[len(common_prefix):]: v for k, v in matches.items()})
         has_descriptions = any(matches.values())
         if has_descriptions or matches.word_transforms:
             lines.append('compdescriptions=(')
-            sz = max(map(wcswidth, matches.transformed_words()))
+            try:
+                sz = max(map(wcswidth, matches.transformed_words()))
+            except ValueError:
+                sz = 0
             limit = min(16, sz)
             for word, desc in matches.transformed_items():
                 lines.extend(map(shlex.quote, fmt_desc(word, desc, limit)))
@@ -384,20 +387,6 @@ def complete_kitty_cli_arg(ans: Completions, opt: Optional[OptionDict], prefix:
         from kitty.config import option_names_for_completion
         k = 'Config directives'
         ans.add_match_group(k, {k+'=': '' for k in option_names_for_completion() if k.startswith(prefix)}, trailing_space=False)
-    elif dest == 'config':
-
-        def is_conf_file(x: str) -> bool:
-            if os.path.isdir(x):
-                return True
-            return x.lower().endswith('.conf')
-
-        complete_files_and_dirs(ans, prefix, files_group_name='Config files', predicate=is_conf_file)
-    elif dest == 'session':
-        complete_files_and_dirs(ans, prefix, files_group_name='Session files')
-    elif dest == 'watcher':
-        complete_files_and_dirs(ans, prefix, files_group_name='Watcher files')
-    elif dest == 'directory':
-        complete_dirs(ans, prefix)
     elif dest == 'listen_on':
         if ':' not in prefix:
             k = 'Address type'
@@ -408,6 +397,15 @@ def complete_kitty_cli_arg(ans: Completions, opt: Optional[OptionDict], prefix:
         complete_basic_option_args(ans, opt, prefix)
 
 
+def basic_option_arg_completer(ans: Completions, opt: Optional[OptionDict], prefix: str, unknown_args: Delegate) -> None:
+    prefix = prefix or ''
+    if not opt:
+        if unknown_args.num_of_unknown_args > 0:
+            ans.delegate = unknown_args
+        return
+    complete_basic_option_args(ans, opt, prefix)
+
+
 CompleteArgsFunc = Callable[[Completions, Optional[OptionDict], str, Delegate], None]
 
 
@@ -416,7 +414,7 @@ def complete_alias_map(
     words: Sequence[str],
     new_word: bool,
     option_map: Dict[str, OptionDict],
-    complete_args: Optional[CompleteArgsFunc] = None
+    complete_args: CompleteArgsFunc = basic_option_arg_completer,
 ) -> None:
     expecting_arg = False
     opt: Optional[OptionDict] = None
@@ -428,11 +426,11 @@ def complete_alias_map(
                 if w is not last_word:
                     continue
                 long_opt = option_map.get(prev_word)
-                if long_opt is not None and complete_args is not None:
+                if long_opt is not None:
                     complete_args(ans, long_opt, '', Delegate())
                     return
             if w is last_word and not new_word:
-                if opt is not None and complete_args is not None:
+                if opt is not None:
                     complete_args(ans, opt, w, Delegate())
                 return
             expecting_arg = False
@@ -441,7 +439,7 @@ def complete_alias_map(
             parts = w.split('=', 1)
             if len(parts) == 2:
                 long_opt = option_map.get(parts[0])
-                if long_opt is not None and complete_args is not None:
+                if long_opt is not None:
                     complete_args(ans, long_opt, parts[1], Delegate())
                     ans.add_prefix(f'{parts[0]}=')
                 return
@@ -450,23 +448,20 @@ def complete_alias_map(
             if w.startswith('-'):
                 ans.add_match_group('Options', {k: opt['help'] for k, opt in option_map.items() if k.startswith(last_word)})
             else:
-                if complete_args is not None:
-                    complete_args(ans, None, last_word, Delegate(words, i))
+                complete_args(ans, None, last_word, Delegate(words, i))
             return
         if opt is None:
-            if complete_args is not None:
-                complete_args(ans, None, '' if new_word else last_word, Delegate(words, i, new_word))
+            complete_args(ans, None, '' if new_word else last_word, Delegate(words, i, new_word))
             if w.startswith('--') and '=' in w:
                 continue
             return  # some non-option word encountered
         expecting_arg = not opt.get('type', '').startswith('bool-')
     if expecting_arg:
-        if opt is not None and complete_args is not None:
+        if opt is not None:
             complete_args(ans, opt, '' if new_word else last_word, Delegate())
     else:
         prefix = '' if new_word else last_word
-        if complete_args is not None:
-            complete_args(ans, None, prefix, Delegate())
+        complete_args(ans, None, prefix, Delegate())
         ans.add_match_group('Options', {k: opt['help'] for k, opt in option_map.items() if k.startswith(prefix)})
 
 
@@ -474,23 +469,33 @@ def complete_cli(
     ans: Completions,
     words: Sequence[str],
     new_word: bool,
-    seq: OptionSpecSeq,
-    complete_args: Optional[CompleteArgsFunc] = None
 ) -> None:
     option_map = {}
-    for opt in seq:
+    for opt in options_for_completion():
         if not isinstance(opt, str):
             for alias in opt['aliases']:
                 option_map[alias] = opt
-    complete_alias_map(ans, words, new_word, option_map, complete_args)
+    complete_alias_map(ans, words, new_word, option_map, complete_kitty_cli_arg)
+
+
+def global_options_for_remote_cmd() -> Dict[str, OptionDict]:
+    seq, disabled = parse_option_spec(global_options_spec())
+    ans: Dict[str, OptionDict] = {}
+    for opt in seq:
+        if isinstance(opt, str):
+            continue
+        for alias in opt['aliases']:
+            ans[alias] = opt
+    return ans
 
 
 def complete_remote_command(ans: Completions, cmd_name: str, words: Sequence[str], new_word: bool) -> None:
     aliases, alias_map = options_for_cmd(cmd_name)
-    if not alias_map:
+    try:
+        args_completion = command_for_name(cmd_name).args_completion
+    except KeyError:
         return
-    args_completer: Optional[CompleteArgsFunc] = None
-    args_completion = command_for_name(cmd_name).args_completion
+    args_completer: CompleteArgsFunc = basic_option_arg_completer
     if args_completion:
         if 'files' in args_completion:
             title, matchers = args_completion['files']
@@ -502,6 +507,16 @@ def complete_remote_command(ans: Completions, cmd_name: str, words: Sequence[str
     complete_alias_map(ans, words, new_word, alias_map, complete_args=args_completer)
 
 
+def complete_launch_wrapper(ans: Completions, words: Sequence[str], new_word: bool, allow_files: bool = True) -> None:
+    from kitty.launch import clone_safe_opts
+    aliases, alias_map = options_for_cmd('launch')
+    alias_map = {k: v for k, v in alias_map.items() if v['dest'] in clone_safe_opts()}
+    args_completer: CompleteArgsFunc = basic_option_arg_completer
+    if allow_files:
+        args_completer = remote_files_completer('Files', ('*',))
+    complete_alias_map(ans, words, new_word, alias_map, complete_args=args_completer)
+
+
 def path_completion(prefix: str = '') -> Tuple[List[str], List[str]]:
     prefix = prefix.replace(r'\ ', ' ')
     dirs, files = [], []
@@ -551,9 +566,102 @@ def complete_files_and_dirs(
         ans.add_match_group(files_group_name, files, is_files=True)
 
 
+def filter_files_from_completion_spec(spec: Dict[str, str]) -> Callable[['os.DirEntry[str]', str], bool]:
+
+    if 'ext' in spec:
+        extensions = frozenset(os.extsep + x.lower() for x in spec['ext'].split(','))
+    else:
+        extensions = frozenset()
+
+    if 'mime' in spec:
+        import re
+        from fnmatch import translate
+        mimes = tuple(re.compile(translate(x)) for x in spec['mime'].split(','))
+        from .guess_mime_type import guess_type
+    else:
+        mimes = ()
+
+    if mimes or extensions:
+        def check_file(x: 'os.DirEntry[str]', result: str) -> bool:
+            if extensions:
+                q = result.lower()
+                for ext in extensions:
+                    if q.endswith(ext):
+                        return True
+            if mimes:
+                mq = guess_type(result)
+                if mq:
+                    for mime in mimes:
+                        if mime.match(mq):
+                            return True
+            return False
+    else:
+        def check_file(x: 'os.DirEntry[str]', result: str) -> bool:
+            return True
+
+    return check_file
+
+
+def complete_file_path(ans: Completions, spec: Dict[str, str], prefix: str, only_dirs: bool = False) -> None:
+    prefix = prefix.replace(r'\ ', ' ')
+    relative_to = spec.get('relative', '')
+    if relative_to:
+        if relative_to == 'conf':
+            relative_to = config_dir
+    else:
+        relative_to = os.getcwd()
+    src_dir = relative_to
+    check_against = prefix
+    prefix_result_with = prefix
+    files, dirs = [], []
+    if prefix:
+        expanded_prefix = os.path.expandvars(os.path.expanduser(prefix))
+        check_against = os.path.basename(expanded_prefix)
+        prefix_result_with = os.path.dirname(expanded_prefix).rstrip(os.sep) + os.sep
+        if os.path.isabs(expanded_prefix):
+            src_dir = os.path.dirname(expanded_prefix)
+        elif os.sep in expanded_prefix or (os.altsep and os.altsep in expanded_prefix):
+            src_dir = os.path.join(relative_to, os.path.dirname(expanded_prefix))
+        else:
+            prefix_result_with = ''
+    try:
+        items: Iterable['os.DirEntry[str]'] = os.scandir(src_dir)
+    except OSError:
+        items = ()
+    check_file = filter_files_from_completion_spec(spec)
+    for x in items:
+        if not x.name.startswith(check_against):
+            continue
+        result = prefix_result_with + x.name
+        if x.is_dir():
+            dirs.append(result.rstrip(os.sep) + os.sep)
+        else:
+            if check_file(x, result):
+                files.append(result)
+    if dirs:
+        ans.add_match_group('Directories', dirs, trailing_space=False, is_files=True)
+    if not only_dirs and files:
+        ans.add_match_group(spec.get('group') or 'Files', files, is_files=True)
+
+
+def complete_path(ans: Completions, opt: OptionDict, prefix: str) -> None:
+    spec = opt['completion']
+    t = spec['type']
+    if 'kwds' in spec:
+        kwds = [x for x in spec['kwds'].split(',') if x.startswith(prefix)]
+        if kwds:
+            ans.add_match_group('Keywords', kwds)
+    if t == 'file':
+        complete_file_path(ans, spec, prefix)
+    elif t == 'directory':
+        complete_file_path(ans, spec, prefix, only_dirs=True)
+
+
 def complete_basic_option_args(ans: Completions, opt: OptionDict, prefix: str) -> None:
     if opt['choices']:
         ans.add_match_group(f'Choices for {opt["dest"]}', tuple(k for k in opt['choices'] if k.startswith(prefix)))
+    elif opt['completion'].get('type') in ('file', 'directory'):
+        complete_path(ans, opt, prefix)
 
 
 def complete_dirs(ans: Completions, prefix: str = '') -> None:
@@ -617,6 +725,19 @@ def remote_args_completer(title: str, words: Iterable[str]) -> CompleteArgsFunc:
     return complete_names_for_arg
 
 
+def remote_command_completer(ans: Completions, opt: Optional[OptionDict], prefix: str, unknown_args: Delegate) -> None:
+    if opt is None:
+        words = unknown_args.words[unknown_args.pos:]
+        new_word = unknown_args.new_word
+        if not words or (len(words) == 1 and not new_word):
+            prefix = (words or ('',))[0]
+            ans.add_match_group('Remote control commands', {c: '' for c in remote_control_command_names() if c.startswith(prefix)})
+        else:
+            complete_remote_command(ans, words[0], words[1:], new_word)
+    else:
+        basic_option_arg_completer(ans, opt, prefix, unknown_args)
+
+
 def config_file_predicate(filename: str) -> bool:
     return filename.endswith('.conf')
 
@@ -647,37 +768,38 @@ def complete_kitten(ans: Completions, kitten: str, words: Sequence[str], new_wor
     options = cd['options']()
     seq = parse_option_spec(options)[0]
     option_map = {}
-    if not new_word:
-        for opt in seq:
-            if not isinstance(opt, str):
-                for alias in opt['aliases']:
-                    option_map[alias] = opt
+    for opt in seq:
+        if not isinstance(opt, str):
+            for alias in opt['aliases']:
+                option_map[alias] = opt
     complete_alias_map(ans, words, new_word, option_map, {
         'icat': complete_icat_args,
         'diff': complete_diff_args,
         'themes': complete_themes_args,
-    }.get(kitten))
+    }.get(kitten, basic_option_arg_completer))
 
 
 def find_completions(words: Sequence[str], new_word: bool, entry_points: Iterable[str], namespaced_entry_points: Iterable[str]) -> Completions:
     ans = Completions()
-    if not words or words[0] != 'kitty':
+    if not words:
+        return ans
+    exe = os.path.basename(words[0])
+    if exe in ('edit-in-kitty', 'clone-in-kitty'):
+        complete_launch_wrapper(ans, words[1:], new_word, allow_files=words[0] != 'clone-in-kitty')
+        return ans
+    if exe != 'kitty':
         return ans
     words = words[1:]
     if not words or (len(words) == 1 and not new_word):
         if words and words[0].startswith('--') and '=' in words[0]:
-            complete_cli(ans, words, new_word, options_for_completion(), complete_kitty_cli_arg)
+            complete_cli(ans, words, new_word)
             return ans
         prefix = words[0] if words else ''
         completions_for_first_word(ans, prefix, entry_points, namespaced_entry_points)
         kitty_cli_opts(ans, prefix)
         return ans
     if words[0] == '@':
-        if len(words) == 1 or (len(words) == 2 and not new_word):
-            prefix = words[1] if len(words) > 1 else ''
-            ans.add_match_group('Remote control commands', {c: '' for c in remote_control_command_names() if c.startswith(prefix)})
-        else:
-            complete_remote_command(ans, words[1], words[2:], new_word)
+        complete_alias_map(ans, words[1:], new_word, global_options_for_remote_cmd(), remote_command_completer)
         return ans
     if words[0].startswith('@'):
         if len(words) == 1 and not new_word:
@@ -685,6 +807,7 @@ def find_completions(words: Sequence[str], new_word: bool, entry_points: Iterabl
             ans.add_match_group('Remote control commands', {f'@{c}': '' for c in remote_control_command_names() if c.startswith(prefix)})
         else:
             complete_remote_command(ans, words[0][1:], words[1:], new_word)
+        return ans
     if words[0] == '+':
         if len(words) == 1 or (len(words) == 2 and not new_word):
             prefix = words[1] if len(words) > 1 else ''
@@ -695,23 +818,29 @@ def find_completions(words: Sequence[str], new_word: bool, entry_points: Iterabl
                     ans.add_match_group('Kittens', (k for k in all_kitten_names() if k.startswith('' if len(words) == 2 else words[2])))
                 else:
                     complete_kitten(ans, words[2], words[3:], new_word)
+            elif words[1] == 'open':
+                complete_cli(ans, words[2:], new_word)
         return ans
     if words[0].startswith('+'):
         if len(words) == 1:
             if new_word:
                 if words[0] == '+kitten':
                     ans.add_match_group('Kittens', all_kitten_names())
+                elif words[0] == '+open':
+                    complete_cli(ans, words[1:], new_word)
             else:
                 prefix = words[0]
                 ans.add_match_group('Entry points', (c for c in namespaced_entry_points if c.startswith(prefix)))
         else:
-            if len(words) == 2 and not new_word:
-                ans.add_match_group('Kittens', (k for k in all_kitten_names() if k.startswith(words[1])))
-            else:
-                if words[0] == '+kitten':
+            if words[0] == '+kitten':
+                if len(words) == 2 and not new_word:
+                    ans.add_match_group('Kittens', (k for k in all_kitten_names() if k.startswith(words[1])))
+                else:
                     complete_kitten(ans, words[1], words[2:], new_word)
+            elif words[0] == '+open':
+                complete_cli(ans, words[1:], new_word)
     else:
-        complete_cli(ans, words, new_word, options_for_completion(), complete_kitty_cli_arg)
+        complete_cli(ans, words, new_word)
 
     return ans
 
