@@ -6,7 +6,7 @@ import os
 import pwd
 import sys
 from contextlib import suppress
-from typing import TYPE_CHECKING, Iterable, NamedTuple, Optional, Set
+from typing import TYPE_CHECKING, Any, Iterator, NamedTuple, Optional, Set
 
 from .types import run_once
 
@@ -22,16 +22,39 @@ class Version(NamedTuple):
 
 appname: str = 'kitty'
 kitty_face = '🐱'
-version: Version = Version(0, 24, 4)
+version: Version = Version(0, 26, 5)
 str_version: str = '.'.join(map(str, version))
 _plat = sys.platform.lower()
 is_macos: bool = 'darwin' in _plat
+is_freebsd: bool = 'freebsd' in _plat
+is_running_from_develop: bool = False
+RC_ENCRYPTION_PROTOCOL_VERSION = '1'
 if getattr(sys, 'frozen', False):
-    extensions_dir: str = getattr(sys, 'kitty_extensions_dir')
-    kitty_base_dir = os.path.dirname(extensions_dir)
-    if is_macos:
-        kitty_base_dir = os.path.dirname(os.path.dirname(kitty_base_dir))
-    kitty_base_dir = os.path.join(kitty_base_dir, 'kitty')
+    extensions_dir: str = getattr(sys, 'kitty_run_data')['extensions_dir']
+
+    def get_frozen_base() -> str:
+        global is_running_from_develop
+        try:
+            from bypy_importer import running_in_develop_mode  # type: ignore
+        except ImportError:
+            pass
+        else:
+            is_running_from_develop = running_in_develop_mode()
+
+        if is_running_from_develop:
+            q = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
+            try:
+                if os.path.isdir(q):
+                    return q
+            except OSError:
+                pass
+        ans = os.path.dirname(extensions_dir)
+        if is_macos:
+            ans = os.path.dirname(os.path.dirname(ans))
+        ans = os.path.join(ans, 'kitty')
+        return ans
+    kitty_base_dir = get_frozen_base()
+    del get_frozen_base
 else:
     kitty_base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
     extensions_dir = os.path.join(kitty_base_dir, 'kitty')
@@ -39,7 +62,7 @@ else:
 
 @run_once
 def kitty_exe() -> str:
-    rpath = sys._xoptions.get('bundle_exe_dir')
+    rpath = getattr(sys, 'kitty_run_data').get('bundle_exe_dir')
     if not rpath:
         items = os.environ.get('PATH', '').split(os.pathsep) + [os.path.join(kitty_base_dir, 'kitty', 'launcher')]
         seen: Set[str] = set()
@@ -116,7 +139,27 @@ def cache_dir() -> str:
     return candidate
 
 
-def wakeup() -> None:
+@run_once
+def runtime_dir() -> str:
+    if 'KITTY_RUNTIME_DIRECTORY' in os.environ:
+        candidate = os.path.abspath(os.environ['KITTY_RUNTIME_DIRECTORY'])
+    elif is_macos:
+        from .fast_data_types import user_cache_dir
+        candidate = user_cache_dir()
+    elif 'XDG_RUNTIME_DIR' in os.environ:
+        candidate = os.path.abspath(os.environ['XDG_RUNTIME_DIR'])
+    else:
+        candidate = f'/run/user/{os.geteuid()}'
+        if not os.path.isdir(candidate) or not os.access(candidate, os.X_OK | os.W_OK | os.R_OK):
+            candidate = os.path.join(cache_dir(), 'run')
+    os.makedirs(candidate, exist_ok=True)
+    import stat
+    if stat.S_IMODE(os.stat(candidate).st_mode) != 0o700:
+        os.chmod(candidate, 0o700)
+    return candidate
+
+
+def wakeup_io_loop() -> None:
     from .fast_data_types import get_boss
     b = get_boss()
     if b is not None:
@@ -133,6 +176,9 @@ except KeyError:
     with suppress(Exception):
         print('Failed to read login shell via getpwuid() for current user, falling back to /bin/sh', file=sys.stderr)
     shell_path = '/bin/sh'
+# Keep this short as it is limited to 103 bytes on macOS
+# https://github.com/ansible/ansible/issues/11536#issuecomment-153030743
+ssh_control_master_template = 'kssh-{kitty_pid}-{ssh_placeholder}'
 
 
 def glfw_path(module: str) -> str:
@@ -173,20 +219,74 @@ def running_in_kitty(set_val: Optional[bool] = None) -> bool:
     return bool(getattr(running_in_kitty, 'ans', False))
 
 
-def list_kitty_resources(package: str = 'kitty') -> Iterable[str]:
-    from importlib.resources import contents
-    return contents(package)
+def list_kitty_resources(package: str = 'kitty') -> Iterator[str]:
+    try:
+        if sys.version_info[:2] < (3, 10):
+            raise ImportError('importlib.resources.files() doesnt work with frozen builds on python 3.9')
+        from importlib.resources import files
+    except ImportError:
+        from importlib.resources import contents
+        return iter(contents(package))
+    else:
+        return (path.name for path in files(package).iterdir())
 
 
 def read_kitty_resource(name: str, package_name: str = 'kitty') -> bytes:
-    from importlib.resources import read_binary
-
-    return read_binary(package_name, name)
+    try:
+        if sys.version_info[:2] < (3, 10):
+            raise ImportError('importlib.resources.files() doesnt work with frozen builds on python 3.9')
+        from importlib.resources import files
+    except ImportError:
+        from importlib.resources import read_binary
+        return read_binary(package_name, name)
+    else:
+        return (files(package_name) / name).read_bytes()
 
 
-def website_url(doc_name: str = '') -> str:
+def website_url(doc_name: str = '', website: str = 'https://sw.kovidgoyal.net/kitty/') -> str:
     if doc_name:
-        doc_name = doc_name.rstrip('/')
-        if doc_name:
-            doc_name += '/'
-    return f'https://sw.kovidgoyal.net/kitty/{doc_name}'
+        base, _, frag = doc_name.partition('#')
+        base = base.rstrip('/')
+        if base:
+            base += '/'
+        doc_name = base + (f'#{frag}' if frag else '')
+    return website + doc_name.lstrip('/')
+
+
+handled_signals: Set[int] = set()
+
+
+def clear_handled_signals(*a: Any) -> None:
+    if not handled_signals:
+        return
+    import signal
+    if hasattr(signal, 'pthread_sigmask'):
+        signal.pthread_sigmask(signal.SIG_UNBLOCK, handled_signals)
+    for s in handled_signals:
+        signal.signal(s, signal.SIG_DFL)
+
+
+@run_once
+def local_docs() -> str:
+    d = os.path.dirname
+    base = d(d(kitty_exe()))
+    from_source = getattr(sys, 'kitty_run_data').get('from_source')
+    if is_macos and from_source and '/kitty.app/Contents/' in kitty_exe():
+        base = d(d(d(base)))
+    subdir = os.path.join('doc', 'kitty', 'html')
+    linux_ans = os.path.join(base, 'share', subdir)
+    if getattr(sys, 'frozen', False):
+        if is_macos:
+            return os.path.join(d(d(d(extensions_dir))), subdir)
+        return linux_ans
+    if os.path.isdir(linux_ans):
+        return linux_ans
+    if from_source:
+        sq = os.path.join(d(base), 'docs', '_build', 'html')
+        if os.path.isdir(sq):
+            return sq
+    for candidate in ('/usr', '/usr/local', '/opt/homebrew'):
+        q = os.path.join(candidate, 'share', subdir)
+        if os.path.isdir(q):
+            return q
+    return ''
