@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import ctypes
@@ -7,33 +6,40 @@ import sys
 from functools import partial
 from math import ceil, cos, floor, pi
 from typing import (
-    Any, Callable, Dict, Generator, List, Optional, Tuple, Union, cast
+    TYPE_CHECKING, Any, Callable, Dict, Generator, List, Optional, Tuple,
+    Union, cast
 )
 
-from kitty.config import defaults
 from kitty.constants import is_macos
 from kitty.fast_data_types import (
     Screen, create_test_font_group, get_fallback_font, set_font_data,
     set_options, set_send_sprite_to_gpu, sprite_map_set_limits,
-    test_render_line, test_shape
+    test_render_line, test_shape, NUM_UNDERLINE_STYLES
 )
 from kitty.fonts.box_drawing import (
-    BufType, render_box_char, render_missing_glyph
+    BufType, distribute_dots, render_box_char, render_missing_glyph
 )
-from kitty.options_stub import Options as OptionsStub
+from kitty.options.types import Options, defaults
 from kitty.typing import CoreTextFont, FontConfigPattern
+from kitty.types import _T
 from kitty.utils import log_error
 
 if is_macos:
-    from .core_text import get_font_files as get_font_files_coretext, font_for_family as font_for_family_macos
+    from .core_text import (
+        find_font_features, font_for_family as font_for_family_macos,
+        get_font_files as get_font_files_coretext
+    )
 else:
-    from .fontconfig import get_font_files as get_font_files_fontconfig, font_for_family as font_for_family_fontconfig
+    from .fontconfig import (
+        find_font_features, font_for_family as font_for_family_fontconfig,
+        get_font_files as get_font_files_fontconfig
+    )
 
 FontObject = Union[CoreTextFont, FontConfigPattern]
 current_faces: List[Tuple[FontObject, bool, bool]] = []
 
 
-def get_font_files(opts: OptionsStub) -> Dict[str, Any]:
+def get_font_files(opts: Options) -> Dict[str, Any]:
     if is_macos:
         return get_font_files_coretext(opts)
     return get_font_files_fontconfig(opts)
@@ -45,10 +51,9 @@ def font_for_family(family: str) -> Tuple[FontObject, bool, bool]:
     return font_for_family_fontconfig(family)
 
 
-Range = Tuple[Tuple[int, int], str]
-
-
-def merge_ranges(a: Range, b: Range, priority_map: Dict[Tuple[int, int], int]) -> Generator[Range, None, None]:
+def merge_ranges(
+    a: Tuple[Tuple[int, int], _T], b: Tuple[Tuple[int, int], _T], priority_map: Dict[Tuple[int, int], int]
+) -> Generator[Tuple[Tuple[int, int], _T], None, None]:
     a_start, a_end = a[0]
     b_start, b_end = b[0]
     a_val, b_val = a[1], b[1]
@@ -90,7 +95,7 @@ def merge_ranges(a: Range, b: Range, priority_map: Dict[Tuple[int, int], int]) -
             after_range = ((b_end + 1, a_end), a_val)
             after_range_prio = a_prio
     # check if the before, mid and after ranges can be coalesced
-    ranges: List[Range] = []
+    ranges: List[Tuple[Tuple[int, int], _T]] = []
     priorities: List[int] = []
     for rq, prio in ((before_range, before_range_prio), (mid_range, mid_range_prio), (after_range, after_range_prio)):
         if rq is None:
@@ -112,7 +117,7 @@ def merge_ranges(a: Range, b: Range, priority_map: Dict[Tuple[int, int], int]) -
     yield from ranges
 
 
-def coalesce_symbol_maps(maps: Dict[Tuple[int, int], str]) -> Dict[Tuple[int, int], str]:
+def coalesce_symbol_maps(maps: Dict[Tuple[int, int], _T]) -> Dict[Tuple[int, int], _T]:
     if not maps:
         return maps
     priority_map = {r: i for i, r in enumerate(maps.keys())}
@@ -136,7 +141,7 @@ def coalesce_symbol_maps(maps: Dict[Tuple[int, int], str]) -> Dict[Tuple[int, in
     return dict(ans)
 
 
-def create_symbol_map(opts: OptionsStub) -> Tuple[Tuple[int, int, int], ...]:
+def create_symbol_map(opts: Options) -> Tuple[Tuple[int, int, int], ...]:
     val = coalesce_symbol_maps(opts.symbol_map)
     family_map: Dict[str, int] = {}
     count = 0
@@ -150,6 +155,10 @@ def create_symbol_map(opts: OptionsStub) -> Tuple[Tuple[int, int, int], ...]:
     return sm
 
 
+def create_narrow_symbols(opts: Options) -> Tuple[Tuple[int, int, int], ...]:
+    return tuple((a, b, v) for (a, b), v in coalesce_symbol_maps(opts.narrow_symbols).items())
+
+
 def descriptor_for_idx(idx: int) -> Tuple[FontObject, bool, bool]:
     return current_faces[idx]
 
@@ -174,7 +183,7 @@ def dump_faces(ftypes: List[str], indices: Dict[str, int]) -> None:
             log_error(face_str(face))
 
 
-def set_font_family(opts: Optional[OptionsStub] = None, override_font_size: Optional[float] = None, debug_font_matching: bool = False) -> None:
+def set_font_family(opts: Optional[Options] = None, override_font_size: Optional[float] = None, debug_font_matching: bool = False) -> None:
     global current_faces
     opts = opts or defaults
     sz = override_font_size or opts.font_size
@@ -188,20 +197,29 @@ def set_font_family(opts: Optional[OptionsStub] = None, override_font_size: Opti
             current_faces.append((font_map[k], 'b' in k, 'i' in k))
     before = len(current_faces)
     sm = create_symbol_map(opts)
+    ns = create_narrow_symbols(opts)
     num_symbol_fonts = len(current_faces) - before
+    font_features = {}
+    for face, _, _ in current_faces:
+        font_features[face['postscript_name']] = find_font_features(face['postscript_name'])
+    font_features.update(opts.font_features)
     if debug_font_matching:
         dump_faces(ftypes, indices)
     set_font_data(
         render_box_drawing, prerender_function, descriptor_for_idx,
         indices['bold'], indices['italic'], indices['bi'], num_symbol_fonts,
-        sm, sz, opts.font_features
+        sm, sz, font_features, ns
     )
 
 
-UnderlineCallback = Callable[[ctypes.Array, int, int, int, int], None]
+if TYPE_CHECKING:
+    CBufType = ctypes.Array[ctypes.c_ubyte]
+else:
+    CBufType = None
+UnderlineCallback = Callable[[CBufType, int, int, int, int], None]
 
 
-def add_line(buf: ctypes.Array, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
+def add_line(buf: CBufType, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
     y = position - thickness // 2
     while thickness > 0 and -1 < y < cell_height:
         thickness -= 1
@@ -209,7 +227,7 @@ def add_line(buf: ctypes.Array, cell_width: int, position: int, thickness: int,
         y += 1
 
 
-def add_dline(buf: ctypes.Array, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
+def add_dline(buf: CBufType, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
     a = min(position - thickness, cell_height - 1)
     b = min(position, cell_height - 1)
     top, bottom = min(a, b), max(a, b)
@@ -229,10 +247,18 @@ def add_dline(buf: ctypes.Array, cell_width: int, position: int, thickness: int,
         ctypes.memset(ctypes.addressof(buf) + (cell_width * y), 255, cell_width)
 
 
-def add_curl(buf: ctypes.Array, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
+def add_curl(buf: CBufType, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
     max_x, max_y = cell_width - 1, cell_height - 1
     xfactor = 2.0 * pi / max_x
-    half_height = max(thickness // 2, 1)
+    thickness = max(1, thickness)
+    if thickness < 3:
+        half_height = thickness
+        thickness -= 1
+    elif thickness == 3:
+        half_height = thickness = 2
+    else:
+        half_height = thickness // 2
+        thickness -= 2
 
     def add_intensity(x: int, y: int, val: int) -> None:
         y += position
@@ -240,20 +266,39 @@ def add_curl(buf: ctypes.Array, cell_width: int, position: int, thickness: int,
         idx = cell_width * y + x
         buf[idx] = min(255, buf[idx] + val)
 
-    # Ensure all space at bottom of cell is used
-    if position + half_height < max_y:
-        position += max_y - (position + half_height)
+    # Ensure curve doesn't exceed cell boundary at the bottom
+    position += half_height * 2
     if position + half_height > max_y:
-        position -= position + half_height - max_y
+        position = max_y - half_height
 
     # Use the Wu antialias algorithm to draw the curve
     # cosine waves always have slope <= 1 so are never steep
     for x in range(cell_width):
         y = half_height * cos(x * xfactor)
-        y1, y2 = floor(y), ceil(y)
-        i1 = int(255 * abs(y - y1))
-        add_intensity(x, y1, 255 - i1)
-        add_intensity(x, y2, i1)
+        y1, y2 = floor(y - thickness), ceil(y)
+        i1 = int(255 * abs(y - floor(y)))
+        add_intensity(x, y1, 255 - i1)  # upper bound
+        add_intensity(x, y2, i1)  # lower bound
+        # fill between upper and lower bound
+        for t in range(1, thickness + 1):
+            add_intensity(x, y1 + t, 255)
+
+
+def add_dots(buf: CBufType, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
+    spacing, size = distribute_dots(cell_width, cell_width // (2 * thickness))
+
+    y = 1 + position - thickness // 2
+    for i in range(y, min(y + thickness, cell_height)):
+        for j, s in enumerate(spacing):
+            buf[cell_width * i + j * size + s: cell_width * i + (j + 1) * size + s] = [255] * size
+
+
+def add_dashes(buf: CBufType, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
+    halfspace_width = cell_width // 4
+    y = 1 + position - thickness // 2
+    for i in range(y, min(y + thickness, cell_height)):
+        buf[cell_width * i:cell_width * i + (cell_width - 3 * halfspace_width)] = [255] * (cell_width - 3 * halfspace_width)
+        buf[cell_width * i + 3 * halfspace_width:cell_width * (i + 1)] = [255] * (cell_width - 3 * halfspace_width)
 
 
 def render_special(
@@ -265,9 +310,11 @@ def render_special(
     underline_position: int = 0,
     underline_thickness: int = 0,
     strikethrough_position: int = 0,
-    strikethrough_thickness: int = 0
-) -> ctypes.Array:
-    underline_position = min(underline_position, cell_height - underline_thickness)
+    strikethrough_thickness: int = 0,
+    dpi_x: float = 96.,
+    dpi_y: float = 96.,
+) -> CBufType:
+    underline_position = min(underline_position, cell_height - sum(divmod(underline_thickness, 2)))
     CharTexture = ctypes.c_ubyte * (cell_width * cell_height)
 
     if missing:
@@ -281,14 +328,13 @@ def render_special(
         try:
             f(ans, cell_width, *a)
         except Exception as e:
-            log_error('Failed to render {} at cell_width={} and cell_height={} with error: {}'.format(
-                f.__name__, cell_width, cell_height, e))
+            log_error(f'Failed to render {f.__name__} at cell_width={cell_width} and cell_height={cell_height} with error: {e}')
 
     if underline:
         t = underline_thickness
         if underline > 1:
             t = max(1, min(cell_height - underline_position - 1, t))
-        dl([add_line, add_line, add_dline, add_curl][underline], underline_position, t, cell_height)
+        dl([add_line, add_line, add_dline, add_curl, add_dots, add_dashes][underline], underline_position, t, cell_height)
     if strikethrough:
         dl(add_line, strikethrough_position, strikethrough_thickness, cell_height)
 
@@ -303,7 +349,7 @@ def render_cursor(
     cell_height: int = 0,
     dpi_x: float = 0,
     dpi_y: float = 0
-) -> ctypes.Array:
+) -> CBufType:
     CharTexture = ctypes.c_ubyte * (cell_width * cell_height)
     ans = CharTexture()
 
@@ -347,21 +393,30 @@ def prerender_function(
     cursor_underline_thickness: float,
     dpi_x: float,
     dpi_y: float
-) -> Tuple[Union[int, ctypes.Array], ...]:
+) -> Tuple[Tuple[int, ...], Tuple[CBufType, ...]]:
     # Pre-render the special underline, strikethrough and missing and cursor cells
     f = partial(
         render_special, cell_width=cell_width, cell_height=cell_height, baseline=baseline,
         underline_position=underline_position, underline_thickness=underline_thickness,
-        strikethrough_position=strikethrough_position, strikethrough_thickness=strikethrough_thickness)
+        strikethrough_position=strikethrough_position, strikethrough_thickness=strikethrough_thickness,
+        dpi_x=dpi_x, dpi_y=dpi_y
+    )
     c = partial(
         render_cursor, cursor_beam_thickness=cursor_beam_thickness,
         cursor_underline_thickness=cursor_underline_thickness, cell_width=cell_width,
         cell_height=cell_height, dpi_x=dpi_x, dpi_y=dpi_y)
-    cells = f(1), f(2), f(3), f(0, True), f(missing=True), c(1), c(2), c(3)
-    return tuple(map(ctypes.addressof, cells)) + (cells,)
+    # If you change the mapping of these cells you will need to change
+    # NUM_UNDERLINE_STYLES and BEAM_IDX in shader.c and STRIKE_SPRITE_INDEX in
+    # window.py and MISSING_GLYPH in font.c
+    cells = list(map(f, range(1, NUM_UNDERLINE_STYLES + 1)))  # underline sprites
+    cells.append(f(0, strikethrough=True))  # strikethrough sprite
+    cells.append(f(missing=True))  # missing glyph
+    cells.extend((c(1), c(2), c(3)))  # cursor glyphs
+    tcells = tuple(cells)
+    return tuple(map(ctypes.addressof, tcells)), tcells
 
 
-def render_box_drawing(codepoint: int, cell_width: int, cell_height: int, dpi: float) -> Tuple[int, ctypes.Array]:
+def render_box_drawing(codepoint: int, cell_width: int, cell_height: int, dpi: float) -> Tuple[int, CBufType]:
     CharTexture = ctypes.c_ubyte * (cell_width * cell_height)
     buf = CharTexture()
     render_box_char(
@@ -428,6 +483,7 @@ def shape_string(
 
 def display_bitmap(rgb_data: bytes, width: int, height: int) -> None:
     from tempfile import NamedTemporaryFile
+
     from kittens.icat.main import detect_support, show
     if not hasattr(display_bitmap, 'detected') and not detect_support():
         raise SystemExit('Your terminal does not support the graphics protocol')
@@ -471,7 +527,7 @@ def test_fallback_font(qtext: Optional[str] = None, bold: bool = False, italic:
             try:
                 print(text, f)
             except UnicodeEncodeError:
-                sys.stdout.buffer.write((text + ' %s\n' % f).encode('utf-8'))
+                sys.stdout.buffer.write(f'{text} {f}\n'.encode('utf-8'))
 
 
 def showcase() -> None:
