@@ -2,6 +2,7 @@
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
+import stat
 from contextlib import suppress
 from typing import Optional
 
@@ -20,7 +21,7 @@ known_extensions = {
 }
 
 
-text_mimes = ('application/javascript', 'application/x-sh', 'application/json')
+text_mimes = ('application/javascript', 'application/x-sh', 'application/x-shellscript', 'application/json')
 
 
 def is_rc_file(path: str) -> bool:
@@ -47,7 +48,15 @@ def initialize_mime_database() -> None:
 
 
 def guess_type(path: str, allow_filesystem_access: bool = False) -> Optional[str]:
-    if allow_filesystem_access and is_folder(path):
+    is_dir = is_exe = False
+
+    if allow_filesystem_access:
+        with suppress(OSError):
+            st = os.stat(path)
+            is_dir = bool(stat.S_ISDIR(st.st_mode))
+            is_exe = bool(not is_dir and st.st_mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH) and os.access(path, os.X_OK))
+
+    if is_dir:
         return 'inode/directory'
     from mimetypes import guess_type as stdlib_guess_type
     initialize_mime_database()
@@ -61,6 +70,9 @@ def guess_type(path: str, allow_filesystem_access: bool = False) -> Optional[str
         mt = f'text/{mt.split("/", 1)[-1]}'
     if not mt and is_rc_file(path):
         mt = 'text/plain'
-    if not mt and is_folder(path):
-        mt = 'inode/directory'
+    if not mt:
+        if is_dir:
+            mt = 'inode/directory'  # type: ignore
+        elif is_exe:
+            mt = 'inode/executable'
     return mt
