@@ -17,9 +17,9 @@ if TYPE_CHECKING:
 class CreateMarker(RemoteCommand):
 
     '''
-    match: Which window to create the marker in
-    self: Boolean indicating whether to create marker in the window the command is run in
-    marker_spec: A list or arguments that define the marker specification, for example: ['text', '1', 'ERROR']
+    match/str: Which window to create the marker in
+    self/bool: Boolean indicating whether to create marker in the window the command is run in
+    marker_spec/list.str: A list or arguments that define the marker specification, for example: ['text', '1', 'ERROR']
     '''
 
     short_desc = 'Create a marker that highlights specified text'
@@ -30,20 +30,24 @@ class CreateMarker(RemoteCommand):
     options_spec = MATCH_WINDOW_OPTION + '''\n
 --self
 type=bool-set
-If specified apply marker to the window this command is run in, rather than the active window.
+Apply marker to the window this command is run in, rather than the active window.
 '''
     argspec = 'MARKER SPECIFICATION'
 
     def message_to_kitty(self, global_opts: RCOptions, opts: 'CLIOptions', args: ArgsType) -> PayloadType:
         if len(args) < 2:
             self.fatal('Invalid marker specification: {}'.format(' '.join(args)))
-        parse_marker_spec(args[0], args[1:])
+        try:
+            parse_marker_spec(args[0], args[1:])
+        except Exception as err:
+            self.fatal(f"Failed to parse marker specification {' '.join(args)} with error: {err}")
         return {'match': opts.match, 'self': opts.self, 'marker_spec': args}
 
     def response_from_kitty(self, boss: Boss, window: Optional[Window], payload_get: PayloadGetType) -> ResponseType:
         args = payload_get('marker_spec')
         for window in self.windows_for_match_payload(boss, window, payload_get):
-            window.set_marker(args)
+            if window:
+                window.set_marker(args)
         return None
 
 
