@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 
@@ -17,7 +16,8 @@ if TYPE_CHECKING:
 class FocusTab(RemoteCommand):
 
     '''
-    match: The tab to focus
+    match/str: The tab to focus
+    no_response/bool: Boolean indicating whether to wait for a response
     '''
 
     short_desc = 'Focus the specified tab'
@@ -33,15 +33,14 @@ using this option means that you will not be notified of failures.
     argspec = ''
 
     def message_to_kitty(self, global_opts: RCOptions, opts: 'CLIOptions', args: ArgsType) -> PayloadType:
-        if opts.no_response:
-            global_opts.no_command_response = True
-        return {'match': opts.match}
+        return {'match': opts.match, 'no_response': opts.no_response}
 
     def response_from_kitty(self, boss: Boss, window: Optional[Window], payload_get: PayloadGetType) -> ResponseType:
         for tab in self.tabs_for_match_payload(boss, window, payload_get):
             if tab:
                 boss.set_active_tab(tab)
                 break
+        return None
 
 
 focus_tab = FocusTab()
