@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 
@@ -17,15 +16,15 @@ if TYPE_CHECKING:
 class RemoveMarker(RemoteCommand):
 
     '''
-    match: Which window to remove the marker from
-    self: Boolean indicating whether to detach the window the command is run in
+    match/str: Which window to remove the marker from
+    self/bool: Boolean indicating whether to detach the window the command is run in
     '''
 
     short_desc = 'Remove the currently set marker, if any.'
     options_spec = MATCH_WINDOW_OPTION + '''\n
 --self
 type=bool-set
-If specified apply marker to the window this command is run in, rather than the active window.
+Apply marker to the window this command is run in, rather than the active window.
 '''
     argspec = ''
 
@@ -34,7 +33,9 @@ If specified apply marker to the window this command is run in, rather than the
 
     def response_from_kitty(self, boss: Boss, window: Optional[Window], payload_get: PayloadGetType) -> ResponseType:
         for window in self.windows_for_match_payload(boss, window, payload_get):
-            window.remove_marker()
+            if window:
+                window.remove_marker()
+        return None
 
 
 remove_marker = RemoveMarker()
