@@ -2,18 +2,19 @@
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
+import re
 from functools import lru_cache, partial, wraps
 from string import Formatter as StringFormatter
 from typing import (
-    Any, Callable, Dict, List, NamedTuple, Optional, Sequence, Tuple, Union
+    Any, Callable, Dict, List, NamedTuple, Optional, Sequence, Tuple, Union,
 )
 
 from .borders import Border, BorderColor
 from .config import build_ansi_color_table
 from .constants import config_dir
 from .fast_data_types import (
-    DECAWM, Region, Screen, cell_size_for_window, get_options, pt_to_px,
-    set_tab_bar_render_data, viewport_for_window, Color
+    DECAWM, Color, Region, Screen, cell_size_for_window, get_boss, get_options,
+    pt_to_px, set_tab_bar_render_data, viewport_for_window,
 )
 from .rgb import alpha_blend, color_as_sgr, color_from_int, to_color
 from .types import WindowGeometry, run_once
@@ -25,6 +26,7 @@ class TabBarData(NamedTuple):
     title: str
     is_active: bool
     needs_attention: bool
+    tab_id: int
     num_windows: int
     num_window_groups: int
     layout_name: str
@@ -150,6 +152,10 @@ class SupSub:
 class ExtraData:
     prev_tab: Optional[TabBarData] = None
     next_tab: Optional[TabBarData] = None
+    # true if the draw_tab function is called just for layout. In such cases,
+    # if drawing is expensive the draw_tab function should avoid drawing and
+    # just move the cursor to its final position, as if drawing was performed.
+    for_layout: bool = False
 
 
 def draw_attributed_string(title: str, screen: Screen) -> None:
@@ -172,13 +178,31 @@ def template_has_field(template: str, field: str) -> bool:
     return False
 
 
-def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int) -> None:
+class TabAccessor:
+
+    def __init__(self, tab_id: int):
+        self.tab_id = tab_id
+
+    @property
+    def active_wd(self) -> str:
+        tab = get_boss().tab_for_id(self.tab_id)
+        return (tab.get_cwd_of_active_window() if tab else '') or ''
+
+
+safe_builtins = {
+    'max': max, 'min': min, 'str': str, 'repr': repr, 'abs': abs, 'len': len, 're': re,
+}
+
+
+def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int, max_title_length: int = 0) -> None:
+    ta = TabAccessor(tab.tab_id)
     data = {
         'index': index,
         'layout_name': tab.layout_name,
         'num_windows': tab.num_windows,
         'num_window_groups': tab.num_window_groups,
         'title': tab.title,
+        'tab': ta,
     }
     ColorFormatter.draw_data = draw_data
     ColorFormatter.tab_data = tab
@@ -188,11 +212,13 @@ def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int)
         'num_windows': tab.num_windows,
         'num_window_groups': tab.num_window_groups,
         'title': tab.title,
+        'tab': ta,
         'fmt': Formatter,
         'sup': SupSub(data),
         'sub': SupSub(data, True),
         'bell_symbol': draw_data.bell_on_tab if tab.needs_attention else '',
         'activity_symbol': draw_data.tab_activity_symbol if tab.has_activity_since_last_focus else '',
+        'max_title_length': max_title_length,
     }
     template = draw_data.title_template
     if tab.is_active and draw_data.active_title_template is not None:
@@ -205,7 +231,7 @@ def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int)
     if prefix:
         template = '{fmt.fg.red}' + prefix + '{fmt.fg.tab}' + template
     try:
-        title = eval(compile_template(template), {'__builtins__': {}}, eval_locals)
+        title = eval(compile_template(template), {'__builtins__': safe_builtins}, eval_locals)
     except Exception as e:
         report_template_failure(template, str(e))
         title = tab.title
@@ -217,7 +243,7 @@ DrawTabFunc = Callable[[DrawData, Screen, TabBarData, int, int, int, bool, Extra
 
 def draw_tab_with_slant(
     draw_data: DrawData, screen: Screen, tab: TabBarData,
-    before: int, max_title_length: int, index: int, is_last: bool,
+    before: int, max_tab_length: int, index: int, is_last: bool,
     extra_data: ExtraData
 ) -> int:
     orig_fg = screen.cursor.fg
@@ -232,20 +258,20 @@ def draw_tab_with_slant(
         screen.cursor.bg = tab_bg
         screen.cursor.fg = orig_fg
 
-    max_title_length += 1
-    if max_title_length <= 1:
+    max_tab_length += 1
+    if max_tab_length <= 1:
         screen.draw('…')
-    elif max_title_length == 2:
+    elif max_tab_length == 2:
         screen.draw('…|')
-    elif max_title_length < 6:
+    elif max_tab_length < 6:
         draw_sep(left_sep)
-        screen.draw((' ' if max_title_length == 5 else '') + '…' + (' ' if max_title_length >= 4 else ''))
+        screen.draw((' ' if max_tab_length == 5 else '') + '…' + (' ' if max_tab_length >= 4 else ''))
         draw_sep(right_sep)
     else:
         draw_sep(left_sep)
         screen.draw(' ')
-        draw_title(draw_data, screen, tab, index)
-        extra = screen.cursor.x - before - max_title_length
+        draw_title(draw_data, screen, tab, index, max_tab_length)
+        extra = screen.cursor.x - before - max_tab_length
         if extra >= 0:
             screen.cursor.x -= extra + 3
             screen.draw('…')
@@ -260,15 +286,15 @@ def draw_tab_with_slant(
 
 def draw_tab_with_separator(
     draw_data: DrawData, screen: Screen, tab: TabBarData,
-    before: int, max_title_length: int, index: int, is_last: bool,
+    before: int, max_tab_length: int, index: int, is_last: bool,
     extra_data: ExtraData
 ) -> int:
     if draw_data.leading_spaces:
         screen.draw(' ' * draw_data.leading_spaces)
-    draw_title(draw_data, screen, tab, index)
-    trailing_spaces = min(max_title_length - 1, draw_data.trailing_spaces)
-    max_title_length -= trailing_spaces
-    extra = screen.cursor.x - before - max_title_length
+    draw_title(draw_data, screen, tab, index, max_tab_length)
+    trailing_spaces = min(max_tab_length - 1, draw_data.trailing_spaces)
+    max_tab_length -= trailing_spaces
+    extra = screen.cursor.x - before - max_tab_length
     if extra > 0:
         screen.cursor.x -= extra + 1
         screen.draw('…')
@@ -286,7 +312,7 @@ def draw_tab_with_separator(
 
 def draw_tab_with_fade(
     draw_data: DrawData, screen: Screen, tab: TabBarData,
-    before: int, max_title_length: int, index: int, is_last: bool,
+    before: int, max_tab_length: int, index: int, is_last: bool,
     extra_data: ExtraData
 ) -> int:
     orig_bg = screen.cursor.bg
@@ -296,12 +322,12 @@ def draw_tab_with_fade(
         screen.cursor.bg = bg
         screen.draw(' ')
     screen.cursor.bg = orig_bg
-    draw_title(draw_data, screen, tab, index)
-    extra = screen.cursor.x - before - max_title_length
+    draw_title(draw_data, screen, tab, index, max(0, max_tab_length - 8))
+    extra = screen.cursor.x - before - max_tab_length
     if extra > 0:
         screen.cursor.x = before
-        draw_title(draw_data, screen, tab, index)
-        extra = screen.cursor.x - before - max_title_length
+        draw_title(draw_data, screen, tab, index, max(0, max_tab_length - 4))
+        extra = screen.cursor.x - before - max_tab_length
         if extra > 0:
             screen.cursor.x -= extra + 1
             screen.draw('…')
@@ -325,7 +351,7 @@ powerline_symbols: Dict[PowerlineStyle, Tuple[str, str]] = {
 
 def draw_tab_with_powerline(
     draw_data: DrawData, screen: Screen, tab: TabBarData,
-    before: int, max_title_length: int, index: int, is_last: bool,
+    before: int, max_tab_length: int, index: int, is_last: bool,
     extra_data: ExtraData
 ) -> int:
     tab_bg = screen.cursor.bg
@@ -348,11 +374,11 @@ def draw_tab_with_powerline(
         start_draw = 1
 
     screen.cursor.bg = tab_bg
-    if min_title_length >= max_title_length:
+    if min_title_length >= max_tab_length:
         screen.draw('…')
     else:
-        draw_title(draw_data, screen, tab, index)
-        extra = screen.cursor.x + start_draw - before - max_title_length
+        draw_title(draw_data, screen, tab, index, max_tab_length)
+        extra = screen.cursor.x + start_draw - before - max_tab_length
         if extra > 0 and extra + 1 < screen.cursor.x:
             screen.cursor.x -= extra + 1
             screen.draw('…')
@@ -395,14 +421,14 @@ def load_custom_draw_tab() -> DrawTabFunc:
     @wraps(func)
     def draw_tab(
         draw_data: DrawData, screen: Screen, tab: TabBarData,
-        before: int, max_title_length: int, index: int, is_last: bool,
+        before: int, max_tab_length: int, index: int, is_last: bool,
         extra_data: ExtraData
     ) -> int:
         try:
-            return func(draw_data, screen, tab, before, max_title_length, index, is_last, extra_data)
+            return func(draw_data, screen, tab, before, max_tab_length, index, is_last, extra_data)
         except Exception as e:
             log_error(f'Custom draw tab function failed with error: {e}')
-            return draw_tab_with_fade(draw_data, screen, tab, before, max_title_length, index, is_last, extra_data)
+            return draw_tab_with_fade(draw_data, screen, tab, before, max_tab_length, index, is_last, extra_data)
 
     return draw_tab
 
@@ -414,6 +440,7 @@ class TabBar:
         self.num_tabs = 1
         self.data_buffer_size = 0
         self.blank_rects: Tuple[Border, ...] = ()
+        self.cell_ranges: List[Tuple[int, int]] = []
         self.laid_out_once = False
         self.apply_options()
 
@@ -501,6 +528,10 @@ class TabBar:
         fg = spec.get('inactive_tab_foreground')
         if fg is None:
             fg = color_as_int(opts.inactive_tab_foreground)
+        else:
+            ifg = color_from_int(fg)
+            if ifg is not None:
+                self.draw_data = self.draw_data._replace(inactive_fg=ifg)
         self.screen.color_profile.set_configured_colors(fg, bg)
 
     def update_blank_rects(self, central: Region, tab_bar: Region, vw: int, vh: int) -> None:
@@ -546,31 +577,67 @@ class TabBar:
         if not self.laid_out_once:
             return
         s = self.screen
-        s.cursor.x = 0
-        s.erase_in_line(2, False)
-        max_title_length = max(1, (self.screen.columns // max(1, len(data))) - 1)
-        cr = []
         last_tab = data[-1] if data else None
         ed = ExtraData()
 
-        for i, t in enumerate(data):
+        def draw_tab(i: int, tab: TabBarData, cell_ranges: List[Tuple[int, int]], max_tab_length: int) -> None:
             ed.prev_tab = data[i - 1] if i > 0 else None
             ed.next_tab = data[i + 1] if i + 1 < len(data) else None
             s.cursor.bg = as_rgb(self.draw_data.tab_bg(t))
             s.cursor.fg = as_rgb(self.draw_data.tab_fg(t))
             s.cursor.bold, s.cursor.italic = self.active_font_style if t.is_active else self.inactive_font_style
             before = s.cursor.x
-            end = self.draw_func(self.draw_data, s, t, before, max_title_length, i + 1, t is last_tab, ed)
+            end = self.draw_func(self.draw_data, s, t, before, max_tab_length, i + 1, t is last_tab, ed)
             s.cursor.bg = s.cursor.fg = 0
-            cr.append((before, end))
-            if s.cursor.x > s.columns - max_title_length and t is not last_tab:
+            cell_ranges.append((before, end))
+            if not ed.for_layout and t is not last_tab and s.cursor.x > s.columns - max_tab_lengths[i+1]:
+                # Stop if there is no space for next tab
                 s.cursor.x = s.columns - 2
                 s.cursor.bg = as_rgb(color_as_int(self.draw_data.default_bg))
                 s.cursor.fg = as_rgb(0xff0000)
                 s.draw(' …')
+                raise StopIteration()
+
+        unconstrained_tab_length = max(1, s.columns - 2)
+        ideal_tab_lengths = [i for i in range(len(data))]
+        default_max_tab_length = max(1, (s.columns // max(1, len(data))) - 1)
+        max_tab_lengths = [default_max_tab_length for _ in range(len(data))]
+        active_idx = 0
+        extra = 0
+        ed.for_layout = True
+        for i, t in enumerate(data):
+            s.cursor.x = 0
+            draw_tab(i, t, [], unconstrained_tab_length)
+            ideal_tab_lengths[i] = tl = max(1, s.cursor.x)
+            if t.is_active:
+                active_idx = i
+            if tl < default_max_tab_length:
+                max_tab_lengths[i] = tl
+                extra += default_max_tab_length - tl
+        if extra > 0:
+            if ideal_tab_lengths[active_idx] > max_tab_lengths[active_idx]:
+                d = min(extra, ideal_tab_lengths[active_idx] - max_tab_lengths[active_idx])
+                max_tab_lengths[active_idx] += d
+                extra -= d
+            if extra > 0:
+                over_achievers = tuple(i for i in range(len(data)) if ideal_tab_lengths[i] > max_tab_lengths[i])
+                if over_achievers:
+                    amt_per_over_achiever = extra // len(over_achievers)
+                    if amt_per_over_achiever > 0:
+                        for i in over_achievers:
+                            max_tab_lengths[i] += amt_per_over_achiever
+
+        s.cursor.x = 0
+        s.erase_in_line(2, False)
+        cr: List[Tuple[int, int]] = []
+        ed.for_layout = False
+        for i, t in enumerate(data):
+            try:
+                draw_tab(i, t, cr, max_tab_lengths[i])
+            except StopIteration:
                 break
-        s.erase_in_line(0, False)  # Ensure no long titles bleed after the last tab
         self.cell_ranges = cr
+        s.erase_in_line(0, False)  # Ensure no long titles bleed after the last tab
         self.align()
 
     def align_with_factor(self, factor: int = 1) -> None:
