@@ -1,44 +1,57 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
+import re
 import stat
 import weakref
 from collections import deque
 from contextlib import suppress
-from functools import partial
 from operator import attrgetter
+from time import monotonic
 from typing import (
-    Any, Deque, Dict, Generator, Iterator, List, NamedTuple, Optional, Pattern,
-    Sequence, Tuple, Union, cast
+    Any, Deque, Dict, Generator, Iterable, Iterator, List, NamedTuple, Optional,
+    Sequence, Set, Tuple, Union,
 )
 
-from .borders import Borders
+from .borders import Border, Borders
 from .child import Child
 from .cli_stub import CLIOptions
 from .constants import appname, kitty_exe
 from .fast_data_types import (
-    add_tab, attach_window, detach_window, get_boss, get_options,
-    mark_tab_bar_dirty, next_window_id, remove_tab, remove_window, ring_bell,
-    set_active_tab, set_active_window, swap_tabs, sync_os_window_title
+    GLFW_MOUSE_BUTTON_LEFT, GLFW_MOUSE_BUTTON_MIDDLE, GLFW_PRESS, GLFW_RELEASE, add_tab,
+    attach_window, current_focused_os_window_id, detach_window, get_boss,
+    get_click_interval, get_options, last_focused_os_window_id, mark_tab_bar_dirty,
+    next_window_id, remove_tab, remove_window, ring_bell, set_active_tab,
+    set_active_window, swap_tabs, sync_os_window_title,
 )
-from .layout.base import Layout, Rect
+from .layout.base import Layout
 from .layout.interface import create_layout_object_for, evict_cached_layouts
 from .tab_bar import TabBar, TabBarData
 from .types import ac
 from .typing import EdgeLiteral, SessionTab, SessionType, TypedDict
 from .utils import log_error, platform_window_id, resolved_shell
-from .window import Watchers, Window, WindowDict
+from .window import CwdRequest, Watchers, Window, WindowDict
 from .window_list import WindowList
 
 
+class TabMouseEvent(NamedTuple):
+    button: int
+    modifiers: int
+    action: int
+    at: float
+    tab_idx: Optional[int]
+
+
 class TabDict(TypedDict):
     id: int
     is_focused: bool
+    is_active_tab: bool
     title: str
     layout: str
     layout_state: Dict[str, Any]
+    layout_opts: Dict[str, Any]
+    enabled_layouts: List[str]
     windows: List[WindowDict]
     active_window_history: List[int]
 
@@ -47,24 +60,26 @@ class SpecialWindowInstance(NamedTuple):
     cmd: Optional[List[str]]
     stdin: Optional[bytes]
     override_title: Optional[str]
-    cwd_from: Optional[int]
+    cwd_from: Optional[CwdRequest]
     cwd: Optional[str]
     overlay_for: Optional[int]
     env: Optional[Dict[str, str]]
     watchers: Optional[Watchers]
+    overlay_behind: bool
 
 
 def SpecialWindow(
     cmd: Optional[List[str]],
     stdin: Optional[bytes] = None,
     override_title: Optional[str] = None,
-    cwd_from: Optional[int] = None,
+    cwd_from: Optional[CwdRequest] = None,
     cwd: Optional[str] = None,
     overlay_for: Optional[int] = None,
     env: Optional[Dict[str, str]] = None,
-    watchers: Optional[Watchers] = None
+    watchers: Optional[Watchers] = None,
+    overlay_behind: bool = False
 ) -> SpecialWindowInstance:
-    return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env, watchers)
+    return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env, watchers, overlay_behind)
 
 
 def add_active_id_to_history(items: Deque[int], item_id: int, maxlen: int = 64) -> None:
@@ -77,26 +92,29 @@ def add_active_id_to_history(items: Deque[int], item_id: int, maxlen: int = 64)
 
 class Tab:  # {{{
 
+    active_fg: Optional[int] = None
+    active_bg: Optional[int] = None
+    inactive_fg: Optional[int] = None
+    inactive_bg: Optional[int] = None
+
     def __init__(
         self,
         tab_manager: 'TabManager',
         session_tab: Optional['SessionTab'] = None,
         special_window: Optional[SpecialWindowInstance] = None,
-        cwd_from: Optional[int] = None,
+        cwd_from: Optional[CwdRequest] = None,
         no_initial_window: bool = False
     ):
         self.tab_manager_ref = weakref.ref(tab_manager)
         self.os_window_id: int = tab_manager.os_window_id
         self.id: int = add_tab(self.os_window_id)
         if not self.id:
-            raise Exception('No OS window with id {} found, or tab counter has wrapped'.format(self.os_window_id))
+            raise Exception(f'No OS window with id {self.os_window_id} found, or tab counter has wrapped')
         self.args = tab_manager.args
         self.name = getattr(session_tab, 'name', '')
         self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or get_options().enabled_layouts]
         self.borders = Borders(self.os_window_id, self.id)
         self.windows = WindowList(self)
-        for i, which in enumerate('first second third fourth fifth sixth seventh eighth ninth tenth'.split()):
-            setattr(self, which + '_window', partial(self.nth_window, num=i))
         self._last_used_layout: Optional[str] = None
         self._current_layout_name: Optional[str] = None
         self.cwd = self.args.directory
@@ -116,14 +134,17 @@ class Tab:  # {{{
             self._set_current_layout(l0)
             self.startup(session_tab)
 
-    def apply_options(self) -> None:
-        for window in self:
-            window.apply_options()
-        self.enabled_layouts = [x.lower() for x in get_options().enabled_layouts] or ['tall']
+    def set_enabled_layouts(self, val: Iterable[str]) -> None:
+        self.enabled_layouts = [x.lower() for x in val] or ['tall']
         if self.current_layout.name not in self.enabled_layouts:
             self._set_current_layout(self.enabled_layouts[0])
         self.relayout()
 
+    def apply_options(self) -> None:
+        for window in self:
+            window.apply_options()
+        self.set_enabled_layouts(get_options().enabled_layouts)
+
     def take_over_from(self, other_tab: 'Tab') -> None:
         self.name, self.cwd = other_tab.name, other_tab.cwd
         self.enabled_layouts = list(other_tab.enabled_layouts)
@@ -138,6 +159,7 @@ class Tab:  # {{{
         for window in self.windows:
             window.change_tab(self)
             attach_window(self.os_window_id, self.id, window.id)
+        self.active_window_changed()
         self.relayout()
 
     def _set_current_layout(self, layout_name: str) -> None:
@@ -147,12 +169,16 @@ class Tab:  # {{{
         self.mark_tab_bar_dirty()
 
     def startup(self, session_tab: 'SessionTab') -> None:
-        for cmd in session_tab.windows:
-            if isinstance(cmd, SpecialWindowInstance):
-                self.new_special_window(cmd)
+        for window in session_tab.windows:
+            spec = window.launch_spec
+            if isinstance(spec, SpecialWindowInstance):
+                self.new_special_window(spec)
             else:
                 from .launch import launch
-                launch(get_boss(), cmd.opts, cmd.args, target_tab=self, force_target_tab=True)
+                launch(get_boss(), spec.opts, spec.args, target_tab=self, force_target_tab=True)
+            if window.resize_spec is not None:
+                self.resize_window(*window.resize_spec)
+
         self.windows.set_active_window_group_for(self.windows.all_windows[session_tab.active_window_idx])
 
     def serialize_state(self) -> Dict[str, Any]:
@@ -162,6 +188,9 @@ class Tab:  # {{{
             'window_list': self.windows.serialize_state(),
             'current_layout': self._current_layout_name,
             'last_used_layout': self._last_used_layout,
+            'layout_opts': self.current_layout.layout_opts,
+            'layout_state': self.current_layout.layout_state,
+            'enabled_layouts': self.enabled_layouts,
             'name': self.name,
         }
 
@@ -183,11 +212,28 @@ class Tab:  # {{{
 
     @property
     def active_window_for_cwd(self) -> Optional[Window]:
-        return self.windows.active_group_base
+        return self.windows.active_group_main
 
     @property
     def title(self) -> str:
-        return cast(str, getattr(self.active_window, 'title', appname))
+        w = self.active_window
+        return w.title if w else appname
+
+    @property
+    def effective_title(self) -> str:
+        return self.name or self.title
+
+    @property
+    def number_of_windows_with_running_programs(self) -> int:
+        ans = 0
+        for window in self:
+            if window.has_running_program:
+                ans += 1
+        return ans
+
+    def get_cwd_of_active_window(self, oldest: bool = False) -> Optional[str]:
+        w = self.active_window
+        return w.get_cwd_of_child(oldest) if w else None
 
     def set_title(self, title: str) -> None:
         self.name = title or ''
@@ -214,7 +260,7 @@ class Tab:  # {{{
             ly = self.current_layout
             self.borders(
                 all_windows=self.windows,
-                current_layout=ly, extra_blank_rects=tm.blank_rects,
+                current_layout=ly, tab_bar_rects=tm.tab_bar_rects,
                 draw_window_borders=(ly.needs_window_borders and self.windows.num_visble_groups > 1) or ly.must_draw_borders
             )
             if w is not None:
@@ -254,7 +300,7 @@ class Tab:  # {{{
         if layout_name not in self.enabled_layouts:
             if raise_exception:
                 raise ValueError(layout_name)
-            log_error('Unknown or disabled layout: {}'.format(layout_name))
+            log_error(f'Unknown or disabled layout: {layout_name}')
             return
         self._set_current_layout(layout_name)
         self.relayout()
@@ -275,7 +321,7 @@ class Tab:  # {{{
             self.goto_layout(layout_name)
 
     def resize_window_by(self, window_id: int, increment: float, is_horizontal: bool) -> Optional[str]:
-        increment_as_percent = self.current_layout.bias_increment_for_cell(is_horizontal) * increment
+        increment_as_percent = self.current_layout.bias_increment_for_cell(self.windows, window_id, is_horizontal) * increment
         if self.current_layout.modify_size_of_window(self.windows, window_id, increment_as_percent, is_horizontal):
             self.relayout()
             return None
@@ -287,6 +333,9 @@ class Tab:  # {{{
         See :ref:`window_resizing` for details.
         ''')
     def resize_window(self, quality: str, increment: int) -> None:
+        if quality == 'reset':
+            self.reset_window_sizes()
+            return
         if increment < 1:
             raise ValueError(increment)
         is_horizontal = quality in ('wider', 'narrower')
@@ -315,10 +364,11 @@ class Tab:  # {{{
         use_shell: bool = False,
         cmd: Optional[List[str]] = None,
         stdin: Optional[bytes] = None,
-        cwd_from: Optional[int] = None,
+        cwd_from: Optional[CwdRequest] = None,
         cwd: Optional[str] = None,
         env: Optional[Dict[str, str]] = None,
-        allow_remote_control: bool = False
+        is_clone_launch: str = '',
+        add_listen_on_env_var: bool = True,
     ) -> Child:
         check_for_suitability = True
         if cmd is None:
@@ -331,12 +381,11 @@ class Tab:  # {{{
                 else:
                     cmd = resolved_shell(get_options())
                     check_for_suitability = False
-                cmd = self.args.args or resolved_shell(get_options())
         if check_for_suitability:
             old_exe = cmd[0]
             if not os.path.isabs(old_exe):
-                import shutil
-                actual_exe = shutil.which(old_exe)
+                from .utils import which
+                actual_exe = which(old_exe)
                 old_exe = actual_exe if actual_exe else os.path.abspath(old_exe)
             try:
                 is_executable = os.access(old_exe, os.X_OK)
@@ -355,12 +404,12 @@ class Tab:  # {{{
                         import shlex
                         with suppress(OSError):
                             with open(old_exe) as f:
-                                cmd = [kitty_exe(), '+hold']
                                 if f.read(2) == '#!':
                                     line = f.read(4096).splitlines()[0]
-                                    cmd += shlex.split(line) + [old_exe]
+                                    cmd[:0] = shlex.split(line)
                                 else:
-                                    cmd += [resolved_shell(get_options())[0], cmd[0]]
+                                    cmd[:0] = [resolved_shell(get_options())[0]]
+                                cmd[:0] = [kitty_exe(), '+hold']
         fenv: Dict[str, str] = {}
         if env:
             fenv.update(env)
@@ -368,12 +417,12 @@ class Tab:  # {{{
         pwid = platform_window_id(self.os_window_id)
         if pwid is not None:
             fenv['WINDOWID'] = str(pwid)
-        ans = Child(cmd, cwd or self.cwd, stdin, fenv, cwd_from, allow_remote_control=allow_remote_control)
+        ans = Child(cmd, cwd or self.cwd, stdin, fenv, cwd_from, is_clone_launch=is_clone_launch, add_listen_on_env_var=add_listen_on_env_var)
         ans.fork()
         return ans
 
-    def _add_window(self, window: Window, location: Optional[str] = None, overlay_for: Optional[int] = None) -> None:
-        self.current_layout.add_window(self.windows, window, location, overlay_for)
+    def _add_window(self, window: Window, location: Optional[str] = None, overlay_for: Optional[int] = None, overlay_behind: bool = False) -> None:
+        self.current_layout.add_window(self.windows, window, location, overlay_for, put_overlay_behind=overlay_behind)
         self.mark_tab_bar_dirty()
         self.relayout()
 
@@ -383,7 +432,7 @@ class Tab:  # {{{
         cmd: Optional[List[str]] = None,
         stdin: Optional[bytes] = None,
         override_title: Optional[str] = None,
-        cwd_from: Optional[int] = None,
+        cwd_from: Optional[CwdRequest] = None,
         cwd: Optional[str] = None,
         overlay_for: Optional[int] = None,
         env: Optional[Dict[str, str]] = None,
@@ -391,17 +440,23 @@ class Tab:  # {{{
         copy_colors_from: Optional[Window] = None,
         allow_remote_control: bool = False,
         marker: Optional[str] = None,
-        watchers: Optional[Watchers] = None
+        watchers: Optional[Watchers] = None,
+        overlay_behind: bool = False,
+        is_clone_launch: str = '',
+        remote_control_passwords: Optional[Dict[str, Sequence[str]]] = None,
     ) -> Window:
         child = self.launch_child(
-            use_shell=use_shell, cmd=cmd, stdin=stdin, cwd_from=cwd_from, cwd=cwd, env=env, allow_remote_control=allow_remote_control)
+            use_shell=use_shell, cmd=cmd, stdin=stdin, cwd_from=cwd_from, cwd=cwd, env=env,
+            is_clone_launch=is_clone_launch, add_listen_on_env_var=False if allow_remote_control and remote_control_passwords else True
+        )
         window = Window(
             self, child, self.args, override_title=override_title,
-            copy_colors_from=copy_colors_from, watchers=watchers
+            copy_colors_from=copy_colors_from, watchers=watchers,
+            allow_remote_control=allow_remote_control, remote_control_passwords=remote_control_passwords
         )
         # Must add child before laying out so that resize_pty succeeds
         get_boss().add_child(window)
-        self._add_window(window, location=location, overlay_for=overlay_for)
+        self._add_window(window, location=location, overlay_for=overlay_for, overlay_behind=overlay_behind)
         if marker:
             try:
                 window.set_marker(marker)
@@ -422,15 +477,9 @@ class Tab:  # {{{
             override_title=special_window.override_title,
             cwd_from=special_window.cwd_from, cwd=special_window.cwd, overlay_for=special_window.overlay_for,
             env=special_window.env, location=location, copy_colors_from=copy_colors_from,
-            allow_remote_control=allow_remote_control, watchers=special_window.watchers
+            allow_remote_control=allow_remote_control, watchers=special_window.watchers, overlay_behind=special_window.overlay_behind
         )
 
-    @ac('win', 'Close the currently active window')
-    def close_window(self) -> None:
-        w = self.active_window
-        if w is not None:
-            self.remove_window(w)
-
     @ac('win', 'Close all windows in the tab other than the currently active window')
     def close_other_windows_in_tab(self) -> None:
         if len(self.windows) > 1:
@@ -439,6 +488,9 @@ class Tab:  # {{{
                 if window is not active_window:
                     self.remove_window(window)
 
+    def move_window_to_top_of_group(self, window: Window) -> bool:
+        return self.windows.move_window_to_top_of_group(window)
+
     def remove_window(self, window: Window, destroy: bool = True) -> None:
         self.windows.remove_window(window)
         if destroy:
@@ -463,28 +515,71 @@ class Tab:  # {{{
         attach_window(self.os_window_id, self.id, window.id)
         self._add_window(window)
 
-    def set_active_window(self, x: Union[Window, int]) -> None:
-        self.windows.set_active_window_group_for(x)
+    def set_active_window(self, x: Union[Window, int], for_keep_focus: Optional[Window] = None) -> None:
+        self.windows.set_active_window_group_for(x, for_keep_focus=for_keep_focus)
 
     def get_nth_window(self, n: int) -> Optional[Window]:
         if self.windows:
             return self.current_layout.nth_window(self.windows, n)
+        return None
 
     @ac('win', '''
-        Focus the nth window if positive or the previously active windows if negative
-
-        For example, to ficus the previously active window::
+        Focus the nth window if positive or the previously active windows if negative. When the number is larger
+        than the number of windows focus the last window. For example::
 
+            # focus the previously active window
             map ctrl+p nth_window -1
+            # focus the first window
+            map ctrl+1 nth_window 0
         ''')
     def nth_window(self, num: int = 0) -> None:
         if self.windows:
             if num < 0:
                 self.windows.make_previous_group_active(-num)
-            else:
-                self.current_layout.activate_nth_window(self.windows, num)
+            elif self.windows.num_groups:
+                self.current_layout.activate_nth_window(self.windows, min(num, self.windows.num_groups - 1))
             self.relayout_borders()
 
+    @ac('win', 'Focus the first window')
+    def first_window(self) -> None:
+        self.nth_window(0)
+
+    @ac('win', 'Focus the second window')
+    def second_window(self) -> None:
+        self.nth_window(1)
+
+    @ac('win', 'Focus the third window')
+    def third_window(self) -> None:
+        self.nth_window(2)
+
+    @ac('win', 'Focus the fourth window')
+    def fourth_window(self) -> None:
+        self.nth_window(3)
+
+    @ac('win', 'Focus the fifth window')
+    def fifth_window(self) -> None:
+        self.nth_window(4)
+
+    @ac('win', 'Focus the sixth window')
+    def sixth_window(self) -> None:
+        self.nth_window(5)
+
+    @ac('win', 'Focus the seventh window')
+    def seventh_window(self) -> None:
+        self.nth_window(6)
+
+    @ac('win', 'Focus the eighth window')
+    def eighth_window(self) -> None:
+        self.nth_window(7)
+
+    @ac('win', 'Focus the ninth window')
+    def ninth_window(self) -> None:
+        self.nth_window(8)
+
+    @ac('win', 'Focus the tenth window')
+    def tenth_window(self) -> None:
+        self.nth_window(9)
+
     def _next_window(self, delta: int = 1) -> None:
         if len(self.windows) > 1:
             self.current_layout.next_window(self.windows, delta)
@@ -510,12 +605,20 @@ class Tab:  # {{{
 
         if groups:
             return groups[0]
+        return None
+
+    def nth_active_window_id(self, n: int = 0) -> int:
+        if n <= 0:
+            return self.active_window.id if self.active_window else 0
+        ids = tuple(reversed(self.windows.active_window_history))
+        return ids[min(n - 1, len(ids) - 1)] if ids else 0
 
     def neighboring_group_id(self, which: EdgeLiteral) -> Optional[int]:
         neighbors = self.current_layout.neighbors(self.windows)
         candidates = neighbors.get(which)
         if candidates:
             return self.most_recent_group(candidates)
+        return None
 
     @ac('win', '''
         Focus the neighboring window in the current tab
@@ -548,6 +651,40 @@ class Tab:  # {{{
                 if self.current_layout.move_window_to_group(self.windows, neighbor):
                     self.relayout()
 
+    def swap_active_window_with(self, window_id: int) -> None:
+        group = self.windows.group_for_window(window_id)
+        if group is not None:
+            w = self.active_window
+            if w is not None and w.id != window_id:
+                if self.current_layout.move_window_to_group(self.windows, group.id):
+                    self.relayout()
+
+    @property
+    def all_window_ids_except_active_window(self) -> Set[int]:
+        all_window_ids = {w.id for w in self}
+        aw = self.active_window
+        if aw is not None:
+            all_window_ids.discard(aw.id)
+        return all_window_ids
+
+    @ac('win', '''
+        Focus a visible window by pressing the number of the window. Window numbers are displayed
+        over the windows for easy selection in this mode. See :opt:`visual_window_select_characters`.
+        ''')
+    def focus_visible_window(self) -> None:
+        def callback(tab: Optional[Tab], window: Optional[Window]) -> None:
+            if tab and window:
+                tab.set_active_window(window)
+
+        get_boss().visual_window_select_action(self, callback, 'Choose window to switch to', only_window_ids=self.all_window_ids_except_active_window)
+
+    @ac('win', 'Swap the current window with another window in the current tab, selected visually. See :opt:`visual_window_select_characters`')
+    def swap_with_window(self) -> None:
+        def callback(tab: Optional[Tab], window: Optional[Window]) -> None:
+            if tab and window:
+                tab.swap_active_window_with(window.id)
+        get_boss().visual_window_select_action(self, callback, 'Choose window to swap with', only_window_ids=self.all_window_ids_except_active_window)
+
     @ac('win', 'Move active window to the top (make it the first window)')
     def move_window_to_top(self) -> None:
         n = self.windows.active_group_idx
@@ -562,15 +699,49 @@ class Tab:  # {{{
     def move_window_backward(self) -> None:
         self.move_window(-1)
 
-    def list_windows(self, active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[WindowDict, None, None]:
+    def list_windows(self, self_window: Optional[Window] = None) -> Generator[WindowDict, None, None]:
+        active_window = self.active_window
         for w in self:
-            yield w.as_dict(is_focused=w is active_window, is_self=w is self_window)
+            yield w.as_dict(
+                is_active_window=w is active_window,
+                is_focused=w.os_window_id == current_focused_os_window_id() and w is active_window,
+                is_self=w is self_window)
 
-    def matches(self, field: str, pat: Pattern) -> bool:
-        if field == 'id':
-            return bool(pat.pattern == str(self.id))
+    def matches_query(self, field: str, query: str, active_tab_manager: Optional['TabManager'] = None) -> bool:
         if field == 'title':
-            return pat.search(self.name or self.title) is not None
+            return re.search(query, self.effective_title) is not None
+        if field == 'id':
+            return query == str(self.id)
+        if field in ('window_id', 'window_title'):
+            field = field.partition('_')[-1]
+            for w in self:
+                if w.matches_query(field, query):
+                    return True
+            return False
+        if field == 'index':
+            if active_tab_manager and len(active_tab_manager.tabs):
+                idx = (int(query) + len(active_tab_manager.tabs)) % len(active_tab_manager.tabs)
+                return active_tab_manager.tabs[idx] is self
+            return False
+        if field == 'recent':
+            if active_tab_manager and len(active_tab_manager.tabs):
+                return self is active_tab_manager.nth_active_tab(int(query))
+            return False
+        if field == 'state':
+            if query == 'active':
+                tm = self.tab_manager_ref()
+                return tm is not None and self is tm.active_tab
+            if query == 'focused':
+                return active_tab_manager is not None and self is active_tab_manager.active_tab and self.os_window_id == last_focused_os_window_id()
+            if query == 'needs_attention':
+                for w in self:
+                    if w.needs_attention:
+                        return True
+            if query == 'parent_active':
+                return active_tab_manager is not None and self.tab_manager_ref() is active_tab_manager
+            if query == 'parent_focused':
+                return active_tab_manager is not None and self.tab_manager_ref() is active_tab_manager and self.os_window_id == last_focused_os_window_id()
+            return False
         return False
 
     def __iter__(self) -> Iterator[Window]:
@@ -593,7 +764,7 @@ class Tab:  # {{{
         self.windows = WindowList(self)
 
     def __repr__(self) -> str:
-        return 'Tab(title={}, id={})'.format(self.name or self.title, hex(id(self)))
+        return f'Tab(title={self.effective_title}, id={self.id})'
 
     def make_active(self) -> None:
         tm = self.tab_manager_ref()
@@ -607,6 +778,7 @@ class TabManager:  # {{{
     def __init__(self, os_window_id: int, args: CLIOptions, wm_class: str, wm_name: str, startup_session: Optional[SessionType] = None):
         self.os_window_id = os_window_id
         self.wm_class = wm_class
+        self.recent_mouse_events: Deque[TabMouseEvent] = deque()
         self.wm_name = wm_name
         self.last_active_tab_id = None
         self.args = args
@@ -673,8 +845,11 @@ class TabManager:  # {{{
         if visible_before and not self.tab_bar_should_be_visible:
             self.tabbar_visibility_changed()
 
-    def _set_active_tab(self, idx: int) -> None:
-        self.active_tab_idx = idx
+    def _set_active_tab(self, idx: int, store_in_history: bool = True) -> None:
+        if store_in_history:
+            self.active_tab_idx = idx
+        else:
+            self._active_tab_idx = idx
         set_active_tab(self.os_window_id, idx)
 
     def tabbar_visibility_changed(self) -> None:
@@ -709,12 +884,16 @@ class TabManager:  # {{{
             tab.relayout_borders()
         self.mark_tab_bar_dirty()
 
-    def set_active_tab(self, tab: Tab) -> bool:
+    def set_active_tab(self, tab: Tab, for_keep_focus: Optional[Tab] = None) -> bool:
         try:
             idx = self.tabs.index(tab)
         except Exception:
             return False
         self.set_active_tab_idx(idx)
+        h = self.active_tab_history
+        if for_keep_focus and len(h) > 2 and h[-2] == for_keep_focus.id and h[-1] != for_keep_focus.id:
+            h.pop()
+            h.pop()
         return True
 
     def next_tab(self, delta: int = 1) -> None:
@@ -732,6 +911,7 @@ class TabManager:  # {{{
             delta = -1 if loc == 'left' else 1
             idx = (len(self.tabs) + self.active_tab_idx + delta) % len(self.tabs)
             return self.tabs[idx]
+        return None
 
     def goto_tab(self, tab_num: int) -> None:
         if tab_num >= len(self.tabs):
@@ -748,21 +928,31 @@ class TabManager:  # {{{
                     self.set_active_tab_idx(idx)
                     break
 
+    def nth_active_tab(self, n: int = 0) -> Optional[Tab]:
+        if n <= 0:
+            return self.active_tab
+        tab_ids = tuple(reversed(self.active_tab_history))
+        return self.tab_for_id(tab_ids[min(n - 1, len(tab_ids) - 1)]) if tab_ids else None
+
     def __iter__(self) -> Iterator[Tab]:
         return iter(self.tabs)
 
     def __len__(self) -> int:
         return len(self.tabs)
 
-    def list_tabs(self, active_tab: Optional[Tab], active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[TabDict, None, None]:
+    def list_tabs(self, self_window: Optional[Window] = None) -> Generator[TabDict, None, None]:
+        active_tab = self.active_tab
         for tab in self:
             yield {
                 'id': tab.id,
-                'is_focused': tab is active_tab,
+                'is_focused': tab is active_tab and tab.os_window_id == current_focused_os_window_id(),
+                'is_active_tab': tab is active_tab,
                 'title': tab.name or tab.title,
                 'layout': str(tab.current_layout.name),
                 'layout_state': tab.current_layout.layout_state(),
-                'windows': list(tab.list_windows(active_window, self_window)),
+                'layout_opts': tab.current_layout.layout_opts.serialized(),
+                'enabled_layouts': tab.enabled_layouts,
+                'windows': list(tab.list_windows(self_window)),
                 'active_window_history': list(tab.windows.active_window_history),
             }
 
@@ -786,19 +976,27 @@ class TabManager:  # {{{
         t = self.active_tab
         if t is not None:
             return t.active_window
+        return None
+
+    @property
+    def number_of_windows_with_running_programs(self) -> int:
+        count = 0
+        for tab in self:
+            count += tab.number_of_windows_with_running_programs
+        return count
 
     @property
     def number_of_windows(self) -> int:
         count = 0
         for tab in self:
-            for window in tab:
-                count += 1
+            count += len(tab)
         return count
 
     def tab_for_id(self, tab_id: int) -> Optional[Tab]:
         for t in self.tabs:
             if t.id == tab_id:
                 return t
+        return None
 
     def move_tab(self, delta: int = 1) -> None:
         if len(self.tabs) > 1:
@@ -814,7 +1012,7 @@ class TabManager:  # {{{
     def new_tab(
         self,
         special_window: Optional[SpecialWindowInstance] = None,
-        cwd_from: Optional[int] = None,
+        cwd_from: Optional[CwdRequest] = None,
         as_neighbor: bool = False,
         empty_tab: bool = False,
         location: str = 'last'
@@ -822,11 +1020,12 @@ class TabManager:  # {{{
         idx = len(self.tabs)
         orig_active_tab_idx = self.active_tab_idx
         self._add_tab(Tab(self, no_initial_window=True) if empty_tab else Tab(self, special_window=special_window, cwd_from=cwd_from))
-        self._set_active_tab(idx)
         if as_neighbor:
             location = 'after'
         if location == 'neighbor':
             location = 'after'
+        if location == 'default':
+            location = 'last'
         if len(self.tabs) > 1 and location != 'last':
             if location == 'first':
                 desired_idx = 0
@@ -836,40 +1035,45 @@ class TabManager:  # {{{
                 for i in range(idx, desired_idx, -1):
                     self.tabs[i], self.tabs[i-1] = self.tabs[i-1], self.tabs[i]
                     swap_tabs(self.os_window_id, i, i-1)
-                self._set_active_tab(desired_idx)
                 idx = desired_idx
+        self._set_active_tab(idx)
         self.mark_tab_bar_dirty()
         return self.tabs[idx]
 
     def remove(self, tab: Tab) -> None:
         active_tab_before_removal = self.active_tab
+        active_tab_needs_to_change = active_tab_before_removal is tab
         self._remove_tab(tab)
-        active_tab = self.active_tab
-        active_tab_needs_to_change = (active_tab is None and (active_tab_before_removal is None or active_tab_before_removal is tab)) or active_tab is tab
         while True:
             try:
                 self.active_tab_history.remove(tab.id)
             except ValueError:
                 break
 
+        def idx_for_id(tab_id: int) -> int:
+            for idx, qtab in enumerate(self.tabs):
+                if qtab.id == tab_id:
+                    return idx
+            return -1
+
         if active_tab_needs_to_change:
             next_active_tab = -1
-            if get_options().tab_switch_strategy == 'previous':
+            tss = get_options().tab_switch_strategy
+            if tss == 'previous':
                 while self.active_tab_history and next_active_tab < 0:
                     tab_id = self.active_tab_history.pop()
-                    for idx, qtab in enumerate(self.tabs):
-                        if qtab.id == tab_id:
-                            next_active_tab = idx
-                            break
-            elif get_options().tab_switch_strategy == 'left':
+                    next_active_tab = idx_for_id(tab_id)
+            elif tss == 'left':
                 next_active_tab = max(0, self.active_tab_idx - 1)
-            elif get_options().tab_switch_strategy == 'right':
+            elif tss == 'right':
                 next_active_tab = min(self.active_tab_idx, len(self.tabs) - 1)
+            elif tss == 'last':
+                next_active_tab = len(self.tabs) - 1
 
             if next_active_tab < 0:
                 next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
 
-            self._set_active_tab(next_active_tab)
+            self._set_active_tab(next_active_tab, store_in_history=False)
         elif active_tab_before_removal is not None:
             try:
                 idx = self.tabs.index(active_tab_before_removal)
@@ -894,22 +1098,43 @@ class TabManager:  # {{{
                 if w.has_activity_since_last_focus:
                     has_activity_since_last_focus = True
             ans.append(TabBarData(
-                title, t is at, needs_attention,
+                title, t is at, needs_attention, t.id,
                 len(t), t.num_window_groups, t.current_layout.name or '',
-                has_activity_since_last_focus
+                has_activity_since_last_focus, t.active_fg, t.active_bg,
+                t.inactive_fg, t.inactive_bg
             ))
         return ans
 
-    def activate_tab_at(self, x: int, is_double: bool = False) -> None:
+    def handle_click_on_tab(self, x: int, button: int, modifiers: int, action: int) -> None:
         i = self.tab_bar.tab_at(x)
+        now = monotonic()
         if i is None:
-            if is_double:
-                self.new_tab()
+            if button == GLFW_MOUSE_BUTTON_LEFT and action == GLFW_RELEASE and len(self.recent_mouse_events) > 2:
+                ci = get_click_interval()
+                prev, prev2 = self.recent_mouse_events[-1], self.recent_mouse_events[-2]
+                if (
+                    prev.button == button and prev2.button == button and
+                    prev.action == GLFW_PRESS and prev2.action == GLFW_RELEASE and
+                    prev.tab_idx is None and prev2.tab_idx is None and
+                    now - prev.at <= ci and now - prev2.at <= 2 * ci
+                ):  # double click
+                    self.new_tab()
+                    self.recent_mouse_events.clear()
+                    return
         else:
-            self.set_active_tab_idx(i)
+            if action == GLFW_PRESS and button == GLFW_MOUSE_BUTTON_LEFT:
+                self.set_active_tab_idx(i)
+            elif button == GLFW_MOUSE_BUTTON_MIDDLE and action == GLFW_RELEASE and self.recent_mouse_events:
+                p = self.recent_mouse_events[-1]
+                if p.button == button and p.action == GLFW_PRESS and p.tab_idx == i:
+                    tab = self.tabs[i]
+                    get_boss().close_tab(tab)
+        self.recent_mouse_events.append(TabMouseEvent(button, modifiers, action, now, i))
+        if len(self.recent_mouse_events) > 5:
+            self.recent_mouse_events.popleft()
 
     @property
-    def blank_rects(self) -> Tuple[Rect, ...]:
+    def tab_bar_rects(self) -> Tuple[Border, ...]:
         return self.tab_bar.blank_rects if self.tab_bar_should_be_visible else ()
 
     def destroy(self) -> None:
