@@ -7,47 +7,55 @@ import re
 import sys
 import weakref
 from collections import deque
-from enum import IntEnum
-from functools import partial
+from contextlib import suppress
+from enum import Enum, IntEnum, auto
+from functools import lru_cache, partial
 from gettext import gettext as _
 from itertools import chain
 from time import monotonic
 from typing import (
-    TYPE_CHECKING, Any, Callable, Deque, Dict, Iterable, List, NamedTuple,
-    Optional, Pattern, Sequence, Tuple, Union
+    TYPE_CHECKING, Any, Callable, Deque, Dict, Iterable, List, NamedTuple, Optional,
+    Pattern, Sequence, Tuple, Union,
 )
 
 from .child import ProcessDesc
 from .cli_stub import CLIOptions
+from .clipboard import (
+    get_clipboard_string, get_primary_selection, set_clipboard_string,
+    set_primary_selection,
+)
 from .config import build_ansi_color_table
-from .constants import appname, is_macos, wakeup
+from .constants import (
+    appname, clear_handled_signals, config_dir, is_macos, wakeup_io_loop,
+)
 from .fast_data_types import (
-    BGIMAGE_PROGRAM, BLIT_PROGRAM, CELL_BG_PROGRAM, CELL_FG_PROGRAM,
-    CELL_PROGRAM, CELL_SPECIAL_PROGRAM, CURSOR_BEAM, CURSOR_BLOCK,
-    CURSOR_UNDERLINE, DCS, DECORATION, DECORATION_MASK, DIM, GLFW_MOD_CONTROL,
-    GRAPHICS_ALPHA_MASK_PROGRAM, GRAPHICS_PREMULT_PROGRAM, GRAPHICS_PROGRAM,
-    MARK, MARK_MASK, NO_CURSOR_SHAPE, NUM_UNDERLINE_STYLES, OSC, REVERSE,
-    SCROLL_FULL, SCROLL_LINE, SCROLL_PAGE, STRIKETHROUGH, TINT_PROGRAM, Color,
-    KeyEvent, Screen, add_timer, add_window, cell_size_for_window,
-    click_mouse_cmd_output, click_mouse_url, compile_program,
-    current_os_window, encode_key_for_tty, get_boss, get_click_interval,
-    get_clipboard_string, get_options, init_cell_program, mark_os_window_dirty,
-    mouse_selection, move_cursor_to_mouse_if_in_prompt, pt_to_px,
-    set_clipboard_string, set_titlebar_color, set_window_logo,
-    set_window_padding, set_window_render_data, update_ime_position_for_window,
-    update_window_title, update_window_visibility
+    BGIMAGE_PROGRAM, BLIT_PROGRAM, CELL_BG_PROGRAM, CELL_FG_PROGRAM, CELL_PROGRAM,
+    CELL_SPECIAL_PROGRAM, CURSOR_BEAM, CURSOR_BLOCK, CURSOR_UNDERLINE, DCS, DECORATION,
+    DECORATION_MASK, DIM, GLFW_MOD_CONTROL, GRAPHICS_ALPHA_MASK_PROGRAM,
+    GRAPHICS_PREMULT_PROGRAM, GRAPHICS_PROGRAM, MARK, MARK_MASK, NO_CURSOR_SHAPE,
+    NUM_UNDERLINE_STYLES, OSC, REVERSE, SCROLL_FULL, SCROLL_LINE, SCROLL_PAGE,
+    STRIKETHROUGH, TINT_PROGRAM, Color, KeyEvent, Screen, add_timer, add_window,
+    cell_size_for_window, click_mouse_cmd_output, click_mouse_url, compile_program,
+    current_focused_os_window_id, encode_key_for_tty, get_boss, get_click_interval,
+    get_options, init_cell_program, last_focused_os_window_id, mark_os_window_dirty,
+    mouse_selection, move_cursor_to_mouse_if_in_prompt, pt_to_px, set_titlebar_color,
+    set_window_logo, set_window_padding, set_window_render_data,
+    update_ime_position_for_window, update_window_title, update_window_visibility,
+    wakeup_main_loop,
 )
 from .keys import keyboard_mode_name, mod_mask
-from .notify import NotificationCommand, handle_notification_cmd
+from .notify import (
+    NotificationCommand, handle_notification_cmd, sanitize_identifier_pat,
+)
 from .options.types import Options
 from .rgb import to_color
 from .terminfo import get_capabilities
-from .types import MouseEvent, WindowGeometry, ac
+from .types import MouseEvent, OverlayType, WindowGeometry, ac, run_once
 from .typing import BossType, ChildType, EdgeLiteral, TabType, TypedDict
 from .utils import (
-    get_primary_selection, kitty_ansi_sanitizer_pat, load_shaders, log_error,
-    open_cmd, open_url, parse_color_set, resolve_custom_file, sanitize_title,
-    set_primary_selection
+    docs_url, kitty_ansi_sanitizer_pat, load_shaders, log_error, open_cmd, open_url,
+    parse_color_set, path_from_osc7_url, resolve_custom_file, resolved_shell,
+    sanitize_for_bracketed_paste, sanitize_title,
 )
 
 MatchPatternType = Union[Pattern[str], Tuple[Pattern[str], Optional[Pattern[str]]]]
@@ -57,6 +65,54 @@ if TYPE_CHECKING:
     from .file_transmission import FileTransmission
 
 
+class CwdRequestType(Enum):
+    current: int = auto()
+    last_reported: int = auto()
+    oldest: int = auto()
+
+
+class CwdRequest:
+
+    def __init__(self, window: Optional['Window'] = None, request_type: CwdRequestType = CwdRequestType.current) -> None:
+        self.window_id = -1 if window is None else window.id
+        self.request_type = request_type
+
+    def __bool__(self) -> bool:
+        return self.window_id > -1
+
+    @property
+    def window(self) -> Optional['Window']:
+        return get_boss().window_id_map.get(self.window_id)
+
+    @property
+    def cwd_of_child(self) -> str:
+        window = self.window
+        if not window:
+            return ''
+        reported_cwd = path_from_osc7_url(window.screen.last_reported_cwd) if window.screen.last_reported_cwd else ''
+        if reported_cwd and not window.child_is_remote and (self.request_type is CwdRequestType.last_reported or window.at_prompt):
+            return reported_cwd
+        return window.get_cwd_of_child(oldest=self.request_type is CwdRequestType.oldest) or ''
+
+    def modify_argv_for_launch_with_cwd(self, argv: List[str]) -> str:
+        window = self.window
+        if not window:
+            return ''
+        reported_cwd = path_from_osc7_url(window.screen.last_reported_cwd) if window.screen.last_reported_cwd else ''
+        if reported_cwd:
+            # First check if we are running ssh kitten, and trying to open the configured login shell
+            if argv[0] == resolved_shell(get_options())[0]:
+                ssh_kitten_cmdline = window.ssh_kitten_cmdline()
+                if ssh_kitten_cmdline:
+                    from kittens.ssh.utils import set_cwd_in_cmdline
+                    argv[:] = ssh_kitten_cmdline
+                    set_cwd_in_cmdline(reported_cwd, argv)
+                    return ''
+            if not window.child_is_remote and (self.request_type is CwdRequestType.last_reported or window.at_prompt):
+                return reported_cwd
+        return window.get_cwd_of_child(oldest=self.request_type is CwdRequestType.oldest) or ''
+
+
 def process_title_from_child(title: str, is_base64: bool) -> str:
     if is_base64:
         from base64 import standard_b64decode
@@ -67,9 +123,23 @@ def process_title_from_child(title: str, is_base64: bool) -> str:
     return sanitize_title(title)
 
 
+@lru_cache(maxsize=64)
+def compile_match_query(exp: str, is_simple: bool = True) -> MatchPatternType:
+    if is_simple:
+        pat: MatchPatternType = re.compile(exp)
+    else:
+        kp, vp = exp.partition('=')[::2]
+        if vp:
+            pat = re.compile(kp), re.compile(vp)
+        else:
+            pat = re.compile(kp), None
+    return pat
+
+
 class WindowDict(TypedDict):
     id: int
     is_focused: bool
+    is_active_window: bool
     title: str
     pid: Optional[int]
     cwd: str
@@ -102,7 +172,7 @@ class DynamicColor(IntEnum):
 
 
 class CommandOutput(IntEnum):
-    last_run, first_on_screen, last_visited = 0, 1, 2
+    last_run, first_on_screen, last_visited, last_non_empty = 0, 1, 2, 3
 
 
 DYNAMIC_COLOR_CODES = {
@@ -206,7 +276,7 @@ def as_text(
     if add_history:
         pht = pagerhist(screen, as_ansi, add_wrap_markers)
         h: List[str] = [pht] if pht else []
-        screen.historybuf.as_text(h.append, as_ansi, add_wrap_markers)
+        screen.as_text_for_history_buf(h.append, as_ansi, add_wrap_markers)
         if h:
             if not screen.linebuf.is_continued(0):
                 h[-1] += '\n'
@@ -293,6 +363,23 @@ def setup_colors(screen: Screen, opts: Options) -> None:
     )
 
 
+@run_once
+def load_paste_filter() -> Callable[[str], str]:
+    import runpy
+    import traceback
+    try:
+        m = runpy.run_path(os.path.join(config_dir, 'paste-actions.py'))
+        func: Callable[[str], str] = m['filter_paste']
+    except Exception as e:
+        if not isinstance(e, FileNotFoundError):
+            traceback.print_exc()
+            log_error(f'Failed to load paste filter function with error: {e}')
+
+        def func(text: str) -> str:
+            return text
+    return func
+
+
 def text_sanitizer(as_ansi: bool, add_wrap_markers: bool) -> Callable[[str], str]:
     pat = kitty_ansi_sanitizer_pat()
     ansi, wrap_markers = not as_ansi, not add_wrap_markers
@@ -318,11 +405,16 @@ def cmd_output(screen: Screen, which: CommandOutput = CommandOutput.last_run, as
         pht = pagerhist(screen, as_ansi, add_wrap_markers, True)
         if pht:
             lines.insert(0, pht)
+    for i in range(min(len(lines), 3)):
+        x = lines[i]
+        if x.startswith('\x1b]133;C'):
+            lines[i] = x.partition('\\')[-1]
     return ''.join(lines)
 
 
-def process_remote_print(msg: bytes) -> str:
+def process_remote_print(msg: str) -> str:
     from base64 import standard_b64decode
+
     from .cli import green
     text = standard_b64decode(msg).decode('utf-8', 'replace')
     return text.replace('\x1b', green(r'\e')).replace('\a', green(r'\a')).replace('\0', green(r'\0'))
@@ -375,6 +467,9 @@ global_watchers = GlobalWatchers()
 
 class Window:
 
+    window_custom_type: str = ''
+    overlay_type = OverlayType.transient
+
     def __init__(
         self,
         tab: TabType,
@@ -382,7 +477,9 @@ class Window:
         args: CLIOptions,
         override_title: Optional[str] = None,
         copy_colors_from: Optional['Window'] = None,
-        watchers: Optional[Watchers] = None
+        watchers: Optional[Watchers] = None,
+        allow_remote_control: bool = False,
+        remote_control_passwords: Optional[Dict[str, Sequence[str]]] = None,
     ):
         if watchers:
             self.watchers = watchers
@@ -391,12 +488,15 @@ class Window:
             self.watchers = global_watchers().copy()
         self.last_focused_at = 0.
         self.started_at = monotonic()
+        self.current_remote_data: List[str] = []
         self.current_mouse_event_button = 0
         self.current_clipboard_read_ask: Optional[bool] = None
         self.prev_osc99_cmd = NotificationCommand()
         self.actions_on_close: List[Callable[['Window'], None]] = []
+        self.actions_on_focus_change: List[Callable[['Window', bool], None]] = []
         self.actions_on_removal: List[Callable[['Window'], None]] = []
         self.current_marker_spec: Optional[Tuple[str, Union[str, Tuple[Tuple[int, str], ...]]]] = None
+        self.kitten_result_processors: List[Callable[['Window', Any], None]] = []
         self.pty_resized_once = False
         self.last_reported_pty_size = (-1, -1, -1, -1)
         self.needs_attention = False
@@ -404,10 +504,10 @@ class Window:
         self.default_title = os.path.basename(child.argv[0] or appname)
         self.child_title = self.default_title
         self.title_stack: Deque[str] = deque(maxlen=10)
-        self.allow_remote_control = child.allow_remote_control
         self.id: int = add_window(tab.os_window_id, tab.id, self.title)
         self.margin = EdgeWidths()
         self.padding = EdgeWidths()
+        self.kitten_result: Optional[Dict[str, Any]] = None
         if not self.id:
             raise Exception(f'No tab with id: {tab.id} in OS Window: {tab.os_window_id} was found, or the window counter wrapped')
         self.tab_id = tab.id
@@ -426,6 +526,27 @@ class Window:
             self.screen.copy_colors_from(copy_colors_from.screen)
         else:
             setup_colors(self.screen, opts)
+        self.remote_control_passwords = remote_control_passwords
+        self.allow_remote_control = allow_remote_control
+
+    def remote_control_allowed(self, pcmd: Dict[str, Any], extra_data: Dict[str, Any]) -> bool:
+        if not self.allow_remote_control:
+            return False
+        if not self.remote_control_passwords:
+            return True
+        pw = pcmd.get('password', '')
+        auth_items = self.remote_control_passwords.get(pw)
+        if pw == '!':
+            auth_items = None
+        if auth_items is None:
+            if '!' in self.remote_control_passwords:
+                raise PermissionError()
+            return False
+        from .remote_control import password_authorizer
+        pa = password_authorizer(auth_items)
+        if not pa.is_cmd_allowed(pcmd, self, False, extra_data):
+            raise PermissionError()
+        return True
 
     @property
     def file_transmission_control(self) -> 'FileTransmission':
@@ -495,11 +616,12 @@ class Window:
     def __repr__(self) -> str:
         return f'Window(title={self.title}, id={self.id})'
 
-    def as_dict(self, is_focused: bool = False, is_self: bool = False) -> WindowDict:
+    def as_dict(self, is_focused: bool = False, is_self: bool = False, is_active_window: bool = False) -> WindowDict:
         return dict(
             id=self.id,
             is_focused=is_focused,
-            title=self.override_title or self.title,
+            is_active_window=is_active_window,
+            title=self.title,
             pid=self.child.pid,
             cwd=self.child.current_cwd or self.child.cwd,
             cmdline=self.child.cmdline,
@@ -511,7 +633,7 @@ class Window:
         )
 
     def serialize_state(self) -> Dict[str, Any]:
-        return {
+        ans = {
             'version': 1,
             'id': self.id,
             'child_title': self.child_title,
@@ -519,12 +641,18 @@ class Window:
             'default_title': self.default_title,
             'title_stack': list(self.title_stack),
             'allow_remote_control': self.allow_remote_control,
+            'remote_control_passwords': self.remote_control_passwords,
             'cwd': self.child.current_cwd or self.child.cwd,
             'env': self.child.environ,
             'cmdline': self.child.cmdline,
             'margin': self.margin.serialize(),
             'padding': self.padding.serialize(),
         }
+        if self.window_custom_type:
+            ans['window_custom_type'] = self.window_custom_type
+        if self.overlay_type is not OverlayType.transient:
+            ans['overlay_type'] = self.overlay_type.value
+        return ans
 
     @property
     def current_colors(self) -> Dict[str, Optional[int]]:
@@ -552,9 +680,9 @@ class Window:
         assert not isinstance(pat, tuple)
 
         if field in ('id', 'window_id'):
-            return True if pat.pattern == str(self.id) else False
+            return pat.pattern == str(self.id)
         if field == 'pid':
-            return True if pat.pattern == str(self.child.pid) else False
+            return pat.pattern == str(self.child.pid)
         if field == 'title':
             return pat.search(self.override_title or self.title) is not None
         if field in 'cwd':
@@ -566,6 +694,38 @@ class Window:
             return False
         return False
 
+    def matches_query(self, field: str, query: str, active_tab: Optional[TabType] = None) -> bool:
+        if field in ('num', 'recent'):
+            if active_tab is not None:
+                try:
+                    q = int(query)
+                except Exception:
+                    return False
+                with suppress(Exception):
+                    if field == 'num':
+                        return active_tab.get_nth_window(q) is self
+                    return active_tab.nth_active_window_id(q) == self.id
+            return False
+        if field == 'state':
+            if query == 'active':
+                tab = self.tabref()
+                return tab is not None and tab.active_window is self
+            if query == 'focused':
+                return active_tab is not None and self is active_tab.active_window and last_focused_os_window_id() == self.os_window_id
+            if query == 'needs_attention':
+                return self.needs_attention
+            if query == 'parent_active':
+                tab = self.tabref()
+                if tab is not None:
+                    tm = tab.tab_manager_ref()
+                    return tm is not None and tm.active_tab is tab
+                return False
+            if query == 'parent_focused':
+                return active_tab is not None and self.tabref() is active_tab and last_focused_os_window_id() == self.os_window_id
+            return False
+        pat = compile_match_query(query, field != 'env')
+        return self.matches(field, pat)
+
     def set_visible_in_layout(self, val: bool) -> None:
         val = bool(val)
         if val is not self.is_visible_in_layout:
@@ -576,7 +736,8 @@ class Window:
 
     def refresh(self) -> None:
         self.screen.mark_as_dirty()
-        wakeup()
+        wakeup_io_loop()
+        wakeup_main_loop()
 
     def set_geometry(self, new_geometry: WindowGeometry) -> None:
         if self.destroyed:
@@ -637,6 +798,8 @@ class Window:
 
     def write_to_child(self, data: Union[str, bytes]) -> None:
         if data:
+            if isinstance(data, str):
+                data = data.encode('utf-8')
             if get_boss().child_monitor.needs_write(self.id, data) is not True:
                 log_error(f'Failed to write to child {self.id} as it does not exist')
 
@@ -715,34 +878,58 @@ class Window:
             set_clipboard_string(url)
 
     def handle_remote_file(self, netloc: str, remote_path: str) -> None:
+        from kittens.remote_file.main import is_ssh_kitten_sentinel
         from kittens.ssh.main import get_connection_data
-        args = self.child.foreground_cmdline
-        conn_data = get_connection_data(args, self.child.foreground_cwd or self.child.current_cwd or '')
+
+        from .utils import SSHConnectionData
+        args = self.ssh_kitten_cmdline()
+        conn_data: Union[None, List[str], SSHConnectionData] = None
+        if args:
+            ssh_cmdline = sorted(self.child.foreground_processes, key=lambda p: p['pid'])[-1]['cmdline'] or ['']
+            if 'ControlPath=' in ' '.join(ssh_cmdline):
+                idx = ssh_cmdline.index('--')
+                conn_data = [is_ssh_kitten_sentinel] + list(ssh_cmdline[:idx + 2])
         if conn_data is None:
-            get_boss().show_error('Could not handle remote file', 'No SSH connection data found in: {args}')
-            return
+            args = self.child.foreground_cmdline
+            conn_data = get_connection_data(args, self.child.foreground_cwd or self.child.current_cwd or '')
+            if conn_data is None:
+                get_boss().show_error('Could not handle remote file', f'No SSH connection data found in: {args}')
+                return
         get_boss().run_kitten(
             'remote_file', '--hostname', netloc.partition(':')[0], '--path', remote_path,
             '--ssh-connection-data', json.dumps(conn_data)
         )
 
+    def send_signal_for_key(self, key_num: int) -> bool:
+        try:
+            return self.child.send_signal_for_key(key_num)
+        except OSError as err:
+            log_error(f'Failed to send signal for key to child with err: {err}')
+            return False
+
     def focus_changed(self, focused: bool) -> None:
         if self.destroyed:
             return
         call_watchers(weakref.ref(self), 'on_focus_change', {'focused': focused})
+        for c in self.actions_on_focus_change:
+            try:
+                c(self, focused)
+            except Exception:
+                import traceback
+                traceback.print_exc()
         self.screen.focus_changed(focused)
         if focused:
             self.last_focused_at = monotonic()
-            update_ime_position_for_window(self.id)
+            update_ime_position_for_window(self.id, False, 1)
             changed = self.needs_attention
             self.needs_attention = False
             if changed:
                 tab = self.tabref()
                 if tab is not None:
                     tab.relayout_borders()
-        elif self.os_window_id == current_os_window():
+        elif self.os_window_id == current_focused_os_window_id():
             # Cancel IME composition after loses focus
-            update_ime_position_for_window(self.id, False, True)
+            update_ime_position_for_window(self.id, False, -1)
 
     def title_changed(self, new_title: Optional[str], is_base64: bool = False) -> None:
         self.child_title = process_title_from_child(new_title or self.default_title, is_base64)
@@ -769,9 +956,9 @@ class Window:
         if cb and cb != ['none']:
             import shlex
             import subprocess
-            env = self.child.final_env
+            env = self.child.foreground_environ
             env['KITTY_CHILD_CMDLINE'] = ' '.join(map(shlex.quote, self.child.cmdline))
-            subprocess.Popen(cb, env=env, cwd=self.child.foreground_cwd)
+            subprocess.Popen(cb, env=env, cwd=self.child.foreground_cwd, preexec_fn=clear_handled_signals)
         if not self.is_active:
             changed = not self.needs_attention
             self.needs_attention = True
@@ -815,6 +1002,10 @@ class Window:
         if default_bg_changed:
             get_boss().default_bg_changed_for(self.id)
 
+    def color_profile_popped(self, bg_changed: bool) -> None:
+        if bg_changed:
+            get_boss().default_bg_changed_for(self.id)
+
     def report_color(self, code: str, r: int, g: int, b: int) -> None:
         r |= r << 8
         g |= g << 8
@@ -822,6 +1013,7 @@ class Window:
         self.screen.send_escape_code_to_child(OSC, f'{code};rgb:{r:04x}/{g:04x}/{b:04x}')
 
     def report_notification_activated(self, identifier: str) -> None:
+        identifier = sanitize_identifier_pat().sub('', identifier)
         self.screen.send_escape_code_to_child(OSC, f'99;i={identifier};')
 
     def set_dynamic_color(self, code: int, value: Union[str, bytes]) -> None:
@@ -875,7 +1067,101 @@ class Window:
     def handle_remote_cmd(self, cmd: str) -> None:
         get_boss().handle_remote_cmd(cmd, self)
 
-    def handle_remote_print(self, msg: bytes) -> None:
+    def handle_remote_echo(self, msg: str) -> None:
+        from base64 import standard_b64decode
+        data = standard_b64decode(msg)
+        self.write_to_child(data)
+
+    def handle_remote_ssh(self, msg: str) -> None:
+        from kittens.ssh.main import get_ssh_data
+        for line in get_ssh_data(msg, f'{os.getpid()}-{self.id}'):
+            self.write_to_child(line)
+
+    def handle_kitten_result(self, msg: str) -> None:
+        import base64
+        self.kitten_result = json.loads(base64.b85decode(msg))
+        for processor in self.kitten_result_processors:
+            try:
+                processor(self, self.kitten_result)
+            except Exception:
+                import traceback
+                traceback.print_exc()
+
+    def add_kitten_result_processor(self, callback: Callable[['Window', Any], None]) -> None:
+        self.kitten_result_processors.append(callback)
+
+    def handle_overlay_ready(self, msg: str) -> None:
+        boss = get_boss()
+        tab = boss.tab_for_window(self)
+        if tab is not None:
+            tab.move_window_to_top_of_group(self)
+
+    def append_remote_data(self, msg: str) -> str:
+        if not msg:
+            cdata = ''.join(self.current_remote_data)
+            self.current_remote_data = []
+            return cdata
+        num, rest = msg.split(':', 1)
+        max_size = get_options().clipboard_max_size * 1024 * 1024
+        if num == '0' or sum(map(len, self.current_remote_data)) > max_size:
+            self.current_remote_data = []
+        self.current_remote_data.append(rest)
+        return ''
+
+    def handle_remote_edit(self, msg: str) -> None:
+        cdata = self.append_remote_data(msg)
+        if cdata:
+            from .launch import remote_edit
+            remote_edit(cdata, self)
+
+    def handle_remote_clone(self, msg: str) -> None:
+        cdata = self.append_remote_data(msg)
+        if cdata:
+            ac = get_options().allow_cloning
+            if ac == 'ask':
+                get_boss().confirm(_(
+                    'A program running in this window wants to clone it into another window.'
+                    ' Allow it do so, once?'),
+                    partial(self.handle_remote_clone_confirmation, cdata), window=self,
+                )
+            elif ac in ('yes', 'y', 'true'):
+                self.handle_remote_clone_confirmation(cdata, True)
+
+    def handle_remote_clone_confirmation(self, cdata: str, confirmed: bool) -> None:
+        if confirmed:
+            from .launch import clone_and_launch
+            clone_and_launch(cdata, self)
+
+    def handle_remote_askpass(self, msg: str) -> None:
+        from .shm import SharedMemory
+        with SharedMemory(name=msg, readonly=True) as shm:
+            shm.seek(1)
+            data = json.loads(shm.read_data_with_size())
+
+        def callback(ans: Any) -> None:
+            data = json.dumps(ans)
+            with SharedMemory(name=msg) as shm:
+                shm.seek(1)
+                shm.write_data_with_size(data)
+                shm.flush()
+                shm.seek(0)
+                shm.write(b'\x01')
+
+        message: str = data['message']
+        if data['type'] == 'confirm':
+            get_boss().confirm(
+                message, callback, window=self, confirm_on_cancel=bool(data.get('confirm_on_cancel')),
+                confirm_on_accept=bool(data.get('confirm_on_accept', True)))
+        elif data['type'] == 'choose':
+            get_boss().choose(
+                message, callback, *data['choices'], window=self, default=data.get('default', ''))
+        elif data['type'] == 'get_line':
+            get_boss().get_line(
+                message, callback, window=self, is_password=bool(data.get('is_password')), prompt=data.get('prompt', '> '))
+        else:
+            log_error(f'Ignoring ask request with unknown type: {data["type"]}')
+
+    def handle_remote_print(self, msg: str) -> None:
         text = process_remote_print(msg)
         print(text, end='', file=sys.stderr)
         sys.stderr.flush()
@@ -1030,13 +1316,13 @@ class Window:
     def paste_selection(self) -> None:
         txt = get_boss().current_primary_selection()
         if txt:
-            self.paste(txt)
+            self.paste_with_actions(txt)
 
     @ac('mouse', 'Paste the current primary selection or the clipboard if no selection is present')
     def paste_selection_or_clipboard(self) -> None:
         txt = get_boss().current_primary_selection_or_clipboard()
         if txt:
-            self.paste(txt)
+            self.paste_with_actions(txt)
 
     @ac('mouse', '''
         Select clicked command output
@@ -1074,10 +1360,11 @@ class Window:
     def destroy(self) -> None:
         self.call_watchers(self.watchers.on_close, {})
         self.destroyed = True
+        del self.kitten_result_processors
         if hasattr(self, 'screen'):
-            if self.is_active and self.os_window_id == current_os_window():
+            if self.is_active and self.os_window_id == current_focused_os_window_id():
                 # Cancel IME composition when window is destroyed
-                update_ime_position_for_window(self.id, False, True)
+                update_ime_position_for_window(self.id, False, -1)
             # Remove cycles so that screen is de-allocated immediately
             self.screen.reset_callbacks()
             del self.screen
@@ -1095,9 +1382,28 @@ class Window:
     def cmd_output(self, which: CommandOutput = CommandOutput.last_run, as_ansi: bool = False, add_wrap_markers: bool = False) -> str:
         return cmd_output(self.screen, which, as_ansi, add_wrap_markers)
 
+    def get_cwd_of_child(self, oldest: bool = False) -> Optional[str]:
+        return self.child.get_foreground_cwd(oldest) or self.child.current_cwd
+
     @property
     def cwd_of_child(self) -> Optional[str]:
-        return self.child.foreground_cwd or self.child.current_cwd
+        return self.get_cwd_of_child()
+
+    @property
+    def child_is_remote(self) -> bool:
+        for p in self.child.foreground_processes:
+            q = list(p['cmdline'] or ())
+            if q and q[0].lower() == 'ssh':
+                return True
+        return False
+
+    def ssh_kitten_cmdline(self) -> List[str]:
+        from kittens.ssh.utils import is_kitten_cmdline
+        for p in self.child.foreground_processes:
+            q = list(p['cmdline'] or ())
+            if is_kitten_cmdline(q):
+                return q
+        return []
 
     def pipe_data(self, text: str, has_wrap_markers: bool = False) -> PipeData:
         text = text or ''
@@ -1119,6 +1425,64 @@ class Window:
         path = resolve_custom_file(path) if path else ''
         set_window_logo(self.os_window_id, self.tab_id, self.id, path, position or '', alpha)
 
+    def paste_with_actions(self, text: str) -> None:
+        if self.destroyed or not text:
+            return
+        opts = get_options()
+        if 'filter' in opts.paste_actions:
+            text = load_paste_filter()(text)
+            if not text:
+                return
+        if 'quote-urls-at-prompt' in opts.paste_actions and self.at_prompt:
+            prefixes = '|'.join(opts.url_prefixes)
+            m = re.match(f'({prefixes}):(.+)', text)
+            if m is not None:
+                scheme, rest = m.group(1), m.group(2)
+                if rest.startswith('//') or scheme in ('mailto', 'irc'):
+                    import shlex
+                    text = shlex.quote(text)
+        btext = text.encode('utf-8')
+        if 'confirm' in opts.paste_actions:
+            msg = ''
+            limit = 16 * 1024
+            if not self.screen.in_bracketed_paste_mode:
+                msg = _('Pasting text into shells that do not support bracketed paste can be dangerous.')
+            elif len(btext) > limit:
+                msg = _('Pasting very large amounts of text ({} bytes) can be slow.').format(len(btext))
+            if msg:
+                get_boss().confirm(msg + _(' Are you sure?'), partial(self.handle_paste_confirmation, btext), window=self)
+                return
+        self.paste_text(btext)
+
+    def handle_paste_confirmation(self, btext: bytes, confirmed: bool) -> None:
+        if confirmed:
+            self.paste_text(btext)
+
+    def paste_bytes(self, text: Union[str, bytes]) -> None:
+        # paste raw bytes without any processing
+        if isinstance(text, str):
+            text = text.encode('utf-8')
+        self.screen.paste_bytes(text)
+
+    def paste_text(self, text: Union[str, bytes]) -> None:
+        if text and not self.destroyed:
+            if isinstance(text, str):
+                text = text.encode('utf-8')
+            if self.screen.in_bracketed_paste_mode:
+                text = sanitize_for_bracketed_paste(text)
+            else:
+                # Workaround for broken editors like nano that cannot handle
+                # newlines in pasted text see https://github.com/kovidgoyal/kitty/issues/994
+                text = text.replace(b'\r\n', b'\n').replace(b'\n', b'\r')
+            self.screen.paste(text)
+
+    def clear_screen(self, reset: bool = False, scrollback: bool = False) -> None:
+        self.screen.cursor.x = self.screen.cursor.y = 0
+        if reset:
+            self.screen.reset()
+        else:
+            self.screen.erase_in_display(3 if scrollback else 2, False)
+
     # actions {{{
 
     @ac('cp', 'Show scrollback in a pager like less')
@@ -1158,28 +1522,17 @@ class Window:
     def show_last_visited_command_output(self) -> None:
         self.show_cmd_output(CommandOutput.last_visited, 'Last visited command output')
 
-    def paste_bytes(self, text: Union[str, bytes]) -> None:
-        # paste raw bytes without any processing
-        if isinstance(text, str):
-            text = text.encode('utf-8')
-        self.screen.paste_bytes(text)
+    @ac('cp', '''
+        Show the last non-empty output from a shell command in a pager like less
+
+        Requires :ref:`shell_integration` to work
+        ''')
+    def show_last_non_empty_command_output(self) -> None:
+        self.show_cmd_output(CommandOutput.last_non_empty, 'Last non-empty command output')
 
     @ac('cp', 'Paste the specified text into the current window')
-    def paste(self, text: Union[str, bytes]) -> None:
-        if text and not self.destroyed:
-            if isinstance(text, str):
-                text = text.encode('utf-8')
-            if self.screen.in_bracketed_paste_mode:
-                while True:
-                    new_text = text.replace(b'\033[201~', b'').replace(b'\x9b201~', b'')
-                    if len(text) == len(new_text):
-                        break
-                    text = new_text
-            else:
-                # Workaround for broken editors like nano that cannot handle
-                # newlines in pasted text see https://github.com/kovidgoyal/kitty/issues/994
-                text = text.replace(b'\r\n', b'\n').replace(b'\n', b'\r')
-            self.screen.paste(text)
+    def paste(self, text: str) -> None:
+        self.paste_with_actions(text)
 
     @ac('cp', 'Copy the selected text from the active window to the clipboard')
     def copy_to_clipboard(self) -> None:
@@ -1207,6 +1560,7 @@ class Window:
         if text:
             set_clipboard_string(text)
         else:
+            self.scroll_end()
             self.write_to_child(self.encoded_key(KeyEvent(key=ord('c'), mods=GLFW_MOD_CONTROL)))
 
     @ac('cp', 'Copy the selected text from the active window to the clipboard and clear selection, if no selection, send SIGINT (aka :kbd:`ctrl+c`)')
@@ -1274,7 +1628,16 @@ class Window:
         if self.screen.is_main_linebuf():
             self.screen.scroll_to_prompt(num_of_prompts)
 
-    @ac('sc', 'Scroll prompt to the bottom of the screen, filling in extra lines form the scrollback buffer')
+    @ac('sc', 'Scroll prompt to the top of the screen, filling screen with empty lines')
+    def scroll_prompt_to_top(self, clear_scrollback: bool = False) -> None:
+        if self.screen.is_main_linebuf():
+            self.screen.scroll_until_cursor_prompt()
+            if clear_scrollback:
+                self.screen.clear_scrollback()
+            elif self.screen.scrolled_by > 0:
+                self.screen.scroll(SCROLL_FULL, False)
+
+    @ac('sc', 'Scroll prompt to the bottom of the screen, filling in extra lines from the scrollback buffer')
     def scroll_prompt_to_bottom(self) -> None:
         self.screen.scroll_prompt_to_bottom()
 
@@ -1314,11 +1677,25 @@ class Window:
 
         For example::
 
-            map F1 signal_child SIGTERM
+            map f1 signal_child SIGTERM
         ''')
     def signal_child(self, *signals: int) -> None:
         pid = self.child.pid_for_cwd
         if pid is not None:
             for sig in signals:
                 os.kill(pid, sig)
+
+    @ac('misc', '''
+    Display the specified kitty documentation, preferring a local copy, if found.
+
+    For example::
+
+        # show the config docs
+        map f1 show_kitty_doc conf
+        # show the ssh kitten docs
+        map f1 show_kitty_doc kittens/ssh
+    ''')
+    def show_kitty_doc(self, which: str = '') -> None:
+        url = docs_url(which)
+        get_boss().open_url(url)
     # }}}
